from datetime import datetime
from typing import Optional, Union

import dash_bootstrap_components as dbc
import humanize
from dash import html
from dash.development.base_component import Component


def component(updated_at: Union[datetime, None]) -> Optional[Component]:
    if updated_at:
        delta = datetime.now() - updated_at
        label = humanize.naturaltime(delta)
        badge = dbc.Badge(
            label, title=updated_at.isoformat(), color="light", className="me-1"
        )
    else:
        badge = dbc.Badge("Unknown", color="warning", className="me-1")

    return html.Div(
        [
            html.Small("Last materialization "),
            badge,
        ]
    )



from feast import FeatureService

from amora.feature_store.feature_view import feature_view_for_model
from amora.feature_store.registry import FEATURE_REGISTRY
from amora.models import Model


def feature_view(model: Model) -> Model:
    """
    Generates a Feature View and a Feature Service for the decorated model.
    Models decorated with `@feature_view` must implement the `FeatureViewSourceProtocol`

    """
    fv = feature_view_for_model(model)

    service = FeatureService(
        name=f"amora_fs__{fv.name}",
        features=[fv],
    )

    FEATURE_REGISTRY[fv.name] = (fv, service, model)

    return model



from typing import List

import dash
import dash_bootstrap_components as dbc
from dash import Input, Output, callback, dcc, html
from dash.development.base_component import Component

from amora.dash.components import question_details
from amora.dash.config import Color
from amora.questions import QUESTIONS

dash.register_page(
    __name__, name="Data Questions", fa_icon="fa-circle-question", location="sidebar"
)


def questions_selector() -> Component:
    return dcc.Dropdown(
        id="questions-selector",
        placeholder="🔍 Search or select",
        options=[
            {"label": question.name, "value": question.uid}
            for question in sorted(QUESTIONS, key=lambda q: q.name)
        ],
        multi=True,
        searchable=True,
    )


def layout() -> Component:
    return dbc.Row(
        [
            html.H1("Data Questions"),
            dbc.Alert(
                f"There are {len(QUESTIONS)} questions registered in this project",
                color=Color.info,
                dismissable=True,
            ),
            dbc.Row(questions_selector()),
            dbc.Row(id="selected-questions"),
        ]
    )


@callback(
    Output("selected-questions", "children"),
    Input("questions-selector", "value"),
    prevent_initial_call=True,
)
def select_value(value: List[str]):
    return [
        dbc.Col(question_details.component(question))
        for question in QUESTIONS
        if question.uid in value
    ]



import dash
import dash_bootstrap_components as dbc
from dash import Input, Output, dcc, html
from dash.development.base_component import Component
from dash_extensions import Lottie

from amora.dag import DependencyDAG
from amora.dash.components import dependency_dag, model_details
from amora.dash.components.animation import Lotties
from amora.models import amora_model_for_name, list_models

dash.register_page(
    __name__,
    fa_icon="fa-database",
    location="sidebar",
    name="Data Models",
    path_template="/models/<unique_identifier>",
    path="/models",
)


def models_selector() -> dcc.Dropdown:
    options = [model.unique_name() for (model, _path) in list_models()]

    return dcc.Dropdown(
        options=options,
        id="model-select-dropdown",
        value=None,
        placeholder="Select a model to begin the exploration",
    )


def layout(unique_identifier: str = None) -> Component:
    if unique_identifier:
        return model_details.component(model=amora_model_for_name(unique_identifier))

    return dbc.Row(
        id="models-content",
        children=[
            dbc.Row(html.H1("Data Models")),
            dbc.Row(models_selector()),
            dbc.Row(dependency_dag.component(dag=DependencyDAG.from_target())),
            dbc.Row(
                id="model-details",
                children=[
                    Lottie(
                        options=dict(
                            loop=True,
                            autoplay=True,
                            rendererSettings=dict(preserveAspectRatio="xMidYMid slice"),
                        ),
                        width="50%",
                        url=Lotties.metaverse_explorer,
                    ),
                ],
            ),
        ],
    )


@dash.callback(
    Output("model-details", "children"),
    Input("model-select-dropdown", "value"),
    prevent_initial_call=True,
)
def update_model_details(value: str) -> Component:
    return model_details.component(model=amora_model_for_name(value))



from datetime import date

import pandas as pd
from numpy import nan
from sqlalchemy import (
    ARRAY,
    Float,
    Integer,
    Numeric,
    String,
    cast,
    func,
    literal,
    union_all,
)
from sqlalchemy_bigquery import STRUCT

from amora.feature_store.protocols import FeatureViewSourceProtocol
from amora.models import Model, select
from amora.providers.bigquery import run
from amora.storage import cache


@cache(suffix=lambda model: f"{model.unique_name()}.{date.today()}")
def summarize(model: Model) -> pd.DataFrame:
    return _summarize_columns(model)


def _summarize_columns(model: Model) -> pd.DataFrame:
    stmts = []

    for column in model.__table__.columns:
        is_supported = not isinstance(column.type, (ARRAY, STRUCT))
        is_numeric = isinstance(column.type, (Numeric, Integer, Float))

        if isinstance(model, FeatureViewSourceProtocol):
            is_fv_feature = column.name in (
                c.name for c in model.feature_view_features()
            )
            is_fv_entity = column.name in (
                c.name for c in model.feature_view_entities()
            )
            is_fv_event_timestamp = (
                column.name == model.feature_view_event_timestamp().name
            )

        else:
            is_fv_feature = False
            is_fv_entity = False
            is_fv_event_timestamp = False

        _min = cast(func.min(column), String) if is_supported else literal(None)
        _max = cast(func.max(column), String) if is_supported else literal(None)
        _unique_count = func.count(column.distinct()) if is_supported else literal(None)
        _avg = cast(func.avg(column), String) if is_numeric else literal(None)
        _stddev = func.stddev(column) if is_numeric else literal(None)
        _null_percentage = (
            func.safe_divide(
                (literal(100) * func.countif(column == None)), func.count(0)
            )
            if is_supported
            else literal(None)
        )

        stmt = select(
            literal(column.name).label("column_name"),
            literal(str(column.type)).label("column_type"),
            _min.label("min"),  # type: ignore
            _max.label("max"),  # type: ignore
            _unique_count.label("unique_count"),  # type: ignore
            _avg.label("avg"),  # type: ignore
            _stddev.label("stddev"),  # type: ignore
            _null_percentage.label("null_percentage"),  # type: ignore
            literal(is_fv_feature).label("is_fv_feature"),
            literal(is_fv_entity).label("is_fv_entity"),
            literal(is_fv_event_timestamp).label("is_fv_event_timestamp"),
        )
        stmts.append(stmt)

    all_stmts = union_all(*stmts)

    result = run(all_stmts)
    df = result.to_dataframe()

    return df.replace({nan: None})



from typing import Iterable

import dash
import dash_bootstrap_components as dbc
from dash import html
from dash.development.base_component import Component
from feast import Feature, FeatureView

from amora.dash.components import (
    materialization_badge,
    model_columns,
    model_datatable,
    model_labels,
    model_summary,
)
from amora.feature_store import fs as store
from amora.feature_store.registry import FEATURE_REGISTRY
from amora.models import Model

dash.register_page(
    __name__, fa_icon="fa-shopping-cart", location="sidebar", name="Feature Store"
)


def entities_list_items(entities: Iterable[str]):
    for entity in entities:
        yield dbc.ListGroupItem(entity, color="primary")


def features_list_items(features: Iterable[Feature]):
    for feature in features:
        yield dbc.ListGroupItem(feature.name)


def icon_for_model(model: Model) -> html.I:
    # fixme: What kind of contract should we expect from the model ?
    icon = getattr(model, "feature_view_fa_icon", None)
    icon = icon() if icon else "fa-square-question"
    return html.I(className=f"fa-solid {icon}")


def card_item(model: Model, fv: FeatureView) -> Component:
    return dbc.Card(
        [
            dbc.CardHeader(
                dbc.Row(
                    [
                        dbc.Col(icon_for_model(model)),
                        dbc.Col(
                            html.H4(fv.name, className="card-title"),
                        ),
                    ],
                    justify="between",
                )
            ),
            dbc.CardBody(
                [
                    model_labels.component(model),
                    materialization_badge.component(fv.last_updated_timestamp),
                    dbc.Accordion(
                        [
                            dbc.AccordionItem(
                                model_summary.component(model),
                                title="📈 Summary",
                            ),
                            dbc.AccordionItem(
                                model_columns.component(model),
                                title="📝 Docs",
                            ),
                            dbc.AccordionItem(
                                model_datatable.component(model),
                                title="🍰 Sample dataset",
                            ),
                        ],
                        start_collapsed=True,
                    ),
                ]
            ),
        ],
    )


def layout() -> Component:
    registry_fvs = {
        fv.name: fv for fv in store.registry.list_feature_views(store.project)
    }

    feature_views = html.Div(
        [
            card_item(model=model, fv=registry_fvs.get(fv.name))
            for (fv, fs, model) in list(FEATURE_REGISTRY.values())
        ]
    )
    return html.Div(
        id="feature-store-content",
        children=[
            html.H1("Feature Store"),
            html.Hr(),
            feature_views,
        ],
    )



from dash import dcc, html
from dash.development.base_component import Component

from amora.dashboards import AcceptedValuesFilter


def layout(filter: AcceptedValuesFilter) -> Component:
    return html.Div(
        [
            filter.title,
            dcc.Dropdown(
                id=filter.id,
                options=filter.values,
                value=filter.default or filter.values,
                multi=True,
            ),
        ]
    )



import hashlib
import inspect
from collections import UserDict
from functools import wraps
from typing import Callable, NamedTuple, Union

import pandas as pd

from amora import logger
from amora.config import StorageCacheProviders, settings
from amora.models import MetaData, create_engine

local_engine = create_engine(
    f"sqlite:///{settings.LOCAL_ENGINE_SQLITE_FILE_PATH}",
    echo=settings.LOCAL_ENGINE_ECHO,
)
local_metadata = MetaData(schema=None)


class CacheKey(NamedTuple):
    """
    E.g:

    ```python
    CacheKey(
        func_module="amora.questions",
        func_name="answer_df",
        func_checksum="e6ccc38abffd7081822da108971e9d9c",
        suffix="how_many_data_points_where_acquired",
    )

    "amora.questions.answer_df.e6ccc38abffd7081822da108971e9d9c.how_many_data_points_where_acquired"
    ```

    """

    func_module: str
    func_name: str
    func_checksum: str
    suffix: str

    def __repr__(self):
        return str(self)

    def __str__(self):
        return ".".join(value for value in self if value)


class Cache(UserDict):
    """
    A `pandas.DataFrame` cache provider that uses the local File System or GCS
    """

    file_suffix: str = ".parquet"

    @property
    def type_(self) -> StorageCacheProviders:
        return settings.STORAGE_CACHE_PROVIDER

    def filepath_for_key(self, key: CacheKey) -> str:
        if self.type_ is StorageCacheProviders.gcs:
            blob_name = f"{key}{self.file_suffix}"
            gs_url = f"gs://{settings.STORAGE_GCS_BUCKET_NAME}/{blob_name}"
            return gs_url

        if self.type_ is StorageCacheProviders.local:
            blob_name = f"{key}{self.file_suffix}"
            return settings.STORAGE_LOCAL_CACHE_PATH.joinpath(blob_name).as_posix()

        raise NotImplementedError

    @logger.log_execution()
    def __setitem__(self, key: CacheKey, value: pd.DataFrame):
        value.to_parquet(
            self.filepath_for_key(key), engine=settings.STORAGE_PARQUET_ENGINE
        )

    @logger.log_execution()
    def __getitem__(self, item: CacheKey) -> pd.DataFrame:
        try:
            return pd.read_parquet(self.filepath_for_key(item))
        except FileNotFoundError as e:
            raise KeyError from e


CACHE = Cache()
Cacheable = Callable[..., pd.DataFrame]


def cache(suffix: Union[Callable[..., str], None] = None):
    """
    Caches a `amora.storage.Cacheable` into the provider selected
    at `setting.STORAGE_CACHE_PROVIDER`.

    To disable the cache, set the env var `AMORA_STORAGE_CACHE_ENABLED` to false.

    ```python
    from datetime import datetime
    from time import sleep

    import pandas as pd

    from amora.storage import cache


    @cache()
    def a_slow_to_build_dataframe():
        sleep(3)
        return pd.DataFrame([{"a": 4, "b": 2}])


    t0 = datetime.now()
    print(a_slow_to_build_dataframe())
    print("Uncached execution:", datetime.now() - t0)

    t0 = datetime.now()
    print(a_slow_to_build_dataframe())
    print("Cached execution:", datetime.now() - t0)
    ```

    ```
       a  b
    0  4  2
    Uncached execution: 0:00:03.011570

       a  b
    0  4  2
    Cached execution: 0:00:00.005567
    ```

    If the cached function expects arguments, a `suffix` function should be provided
    to the `cache` decorator. The function will be called with the same arguments a the
    cached function, and a `str` should be returned.

    ```python
    @cache(suffix=lambda arg1, arg2: f"{arg1}_{arg2}")
    def cacheable_func(arg1, arg2):
        return pd.DataFrame([{"amora": arg1, "storage": arg2}])


    cacheable_func(4, 2)
    # CacheKey -> foo_module.cacheable_func.e6ccc38abffd7081822da108971e9d9c.4_2

    cacheable_func(5, 3)
    # CacheKey -> foo_module.cacheable_func.e6ccc38abffd7081822da108971e9d9c.5_3
    ```

    The key function can also be used to implement a cache TTL:

    ```python
    from datetime import date


    @cache(suffix=lambda: date.today().isoformat())
    def im_cached_for_one_day():
        sleep(3)
        return pd.DataFrame([[1, 2, 3], [4, 5, 6]])
    ```
    """

    def wrapper(fn: Cacheable):
        func_checksum = hashlib.md5(inspect.getsource(fn).encode("utf-8")).hexdigest()

        @wraps(fn)
        def decorator(*args, **kwargs):
            if not settings.STORAGE_CACHE_ENABLED:
                return fn(*args, **kwargs)

            cache_key = CacheKey(
                func_module=fn.__module__,
                func_name=fn.__name__,
                func_checksum=func_checksum,
                suffix=suffix(*args, **kwargs) if suffix else "",
            )

            try:
                return CACHE[cache_key]
            except KeyError:
                result = fn(*args, **kwargs)
                CACHE[cache_key] = result
                return result

        return decorator

    return wrapper



import itertools
import json
import os
from typing import Callable, Iterable, Optional

import pytest
from sqlalchemy import ARRAY, Integer, and_, func, literal, or_, union_all
from sqlmodel.sql.expression import SelectOfScalar

from amora.config import settings
from amora.models import AmoraModel, ColumnElement, Session, select
from amora.protocols import Compilable
from amora.providers.bigquery import RunResult, estimated_query_cost_in_usd, run
from amora.storage import local_engine
from amora.tests.audit import AuditLog

Test = Callable[..., SelectOfScalar]


def _log_result(run_result: RunResult) -> AuditLog:
    with Session(local_engine) as session:
        log = AuditLog(
            bytes_billed=run_result.total_bytes,
            estimated_cost_in_usd=estimated_query_cost_in_usd(run_result.total_bytes),
            execution_time_in_ms=run_result.execution_time_in_ms,
            query=run_result.query,
            referenced_tables=json.dumps(run_result.referenced_tables),
            test_node_id=os.getenv("PYTEST_CURRENT_TEST"),
            test_run_id=settings.TEST_RUN_ID,
            user_email=run_result.user_email,
        )
        session.add(log)
        session.commit()
        return log


def _test(statement: Compilable, raise_on_fail: bool = True) -> bool:
    """
    :param statement: A str with a valid SQL compiled statement
    :param raise_on_fail: By default, the test will raise a pytest Fail exception, with a debug message. Default `True`.
    :return: `True` if the test passed, `False` otherwise
    """
    run_result = run(statement)
    _log_result(run_result)

    if run_result.rows.total_rows == 0:
        return True

    if raise_on_fail:
        pytest.fail(
            f"{run_result.rows.total_rows} rows failed the test assertion."
            f"\n==========="
            f"\nTest query:"
            f"\n==========="
            f"\n{run_result.query}",
            pytrace=False,
        )

    return False


def that(
    column: ColumnElement,
    test: Test,
    raise_on_fail: bool = True,
    **test_kwargs,
) -> bool:
    """
    Executes the test, returning `True` if the test is successful and raising a pytest fail otherwise

    Example:

    ```python
    assert that(HeartRate.value, is_not_null)
    ```
    :param column: An AmoraModel column to test
    :param test: The test assertion function
    :param raise_on_fail: By default, the test will raise a pytest Fail exception, with a debug message.  Default `True`.
    :param test_kwargs: Keyword arguments passed to the `test` function

    """
    return _test(statement=test(column, **test_kwargs), raise_on_fail=raise_on_fail)


def is_not_null(column: ColumnElement) -> Compilable:
    """
    Asserts that the `column` does not contain `null` values

    Results in the following query:

    ```sql
    SELECT {{ column_name }}
    FROM {{ model }}
    WHERE {{ column_name }} IS NULL
    ```

    Example:

    ```python
    is_not_null(HeartRate.id)
    ```

    """
    return select(column).where(column == None)


def is_unique(column: ColumnElement) -> Compilable:
    """
    Assert that the `column` values are unique

    Example SQL:

    ```sql
    SELECT {{ column_name }}
    FROM (
        SELECT {{ column_name }}
        FROM {{ model }}
        WHERE {{ column_name }} IS NOT NULL
        GROUP BY {{ column_name }}
        HAVING COUNT(*) > 1
    ) validation_errors
    ```

    Example:

    ```python
    is_unique(HeartRate.id)
    ```
    """
    return select(column).group_by(column).having(func.count(column) > 1)


def has_accepted_values(column: ColumnElement, values: Iterable) -> Compilable:
    """
    Assert that the values from the `column` should be one of the provided `values`

    Example SQL:

    ```sql
    SELECT {{ column_name }}
    FROM {{ model }}
    WHERE {{ column_name }} NOT IN {{ values }}
    ```

    Example:

    ```python
    has_accepted_values(HeartRate.source, values=["iPhone", "Mi Band"])
    ```
    """
    return select(column).where(~column.in_(values))


def relationship(
    from_: ColumnElement,
    to: ColumnElement,
    from_condition=None,
    to_condition=None,
) -> bool:
    """
    Each value of the `from_` column exists as a value in the `to` column.
    Also known as referential integrity.

    This test validates the referential integrity between two relations
    with a predicate (`from_condition` and `to_condition`) to filter out
    some rows from the test. This is useful to exclude records such as
    test entities, rows created in the last X minutes/hours to account
    for temporary gaps due to data ingestion limitations, etc.

    Example SQL:

    ```sql
    WITH left_table AS (
      SELECT
        {{from_column_name}} AS id
      FROM {{from_table}}
      WHERE
        {{from_column_name}} IS NOT NULL
        AND {{from_condition}}
    ),
    right_table AS (
      SELECT
        {{to_column_name}} AS id
      FROM {{to_table}}
      WHERE
        {{to_column_name}} IS NOT NULL
        AND {{to_condition}}
    ),
    exceptions as (
      SELECT
        left_table.id AS {{from_column_name}}}
      FROM
        left_table
      LEFT JOIN
        right_table
        ON left_table.id = right_table.id
      WHERE
        right_table.id IS NULL
    )

    SELECT * FROM exceptions
    ```

    Example:

    ```python
    relationship(HeartRate.id, to=Health.id)
    ```

    """
    left_table = (
        select(from_.label("id"))
        .where(from_ != None)
        .where(from_condition or and_(True))
        .cte("left_table")
    )
    right_table = (
        select(to.label("id"))
        .where(to != None)
        .where(to_condition or and_(True))
        .cte("right_table")
    )

    exceptions = (
        select([left_table.c["id"].label(from_.key)])
        .select_from(
            left_table.join(
                right_table,
                onclause=left_table.c["id"] == right_table.c["id"],
                isouter=True,
            )
        )
        .where(right_table.c["id"] == None)
    )

    return _test(statement=exceptions)


def is_numeric(column: ColumnElement) -> Compilable:
    """
    Asserts that each not null value is a number

    Example SQL:

    ```sql
        SELECT
            {{ column }}
        FROM
            {{ model }}
        WHERE
            REGEXP_CONTAINS({{ column }}, "[^0-9]")
    ```

    Example:

    ```python
    is_numeric(func.cast(Health.value, String).label("value_as_str"))
    ```
    """
    return select(column).where(func.REGEXP_CONTAINS(column, "[^0-9]"))


def is_non_negative(column: ColumnElement) -> Compilable:
    """
    Asserts that every column value should be >= 0

    Example SQL:

    ```sql
    SELECT {{ column_name }}
    FROM {{ model }}
    WHERE {{ column_name }} < 0
    ```

    Example:

    ```python
    is_non_negative(HeartRate.value)
    ```
    """
    return select(column).where(column < 0)


def is_a_non_empty_string(column: ColumnElement) -> Compilable:
    """
    Asserts that the column isn't an empty string

    Example SQL:

    ```sql
    SELECT
         {{ column_name }}
    FROM
         {{ model }}
    WHERE
         TRIM({{ column_name }}) = ""
    ```

    Example:

    ```python
    is_a_non_empty_string(Health.source)
    ```
    """
    return select(column).where(func.trim(column) == literal(""))


def expression_is_true(expression, condition=None) -> bool:
    """
    Asserts that a expression is TRUE for all records.
    This is useful when checking integrity across columns, for example,
    that a total is equal to the sum of its parts, or that at least one column is true.

    Optionally assert `expression` only for rows where `condition` is met.

    Arguments:
        condition (object): A query filter

    Example:

    ```python
    expression_is_true(StepsAgg._sum > StepsAgg._avg, condition=StepsAgg.year == 2021)
    ```

    """
    statement = select(["*"]).where(~expression)

    if condition is not None:
        statement = statement.where(condition)

    return _test(statement)


def equality(
    model_a: AmoraModel,
    model_b: AmoraModel,
    compare_columns: Optional[Iterable[ColumnElement]] = None,
) -> bool:
    """
    This schema test asserts the equality of two models. Optionally specify a subset of columns to compare.

    """

    raise NotImplementedError

    def comparable_columns(model: AmoraModel) -> Iterable[ColumnElement]:
        if not compare_columns:
            return model
        return [getattr(model, column_name) for column_name in compare_columns]

    a = select(comparable_columns(model_a)).cte("a")
    b = select(comparable_columns(model_b)).cte("b")

    # fixme: google.api_core.exceptions.BadRequest: 400 EXCEPT must be followed by ALL, DISTINCT, or "(" at [34:4]
    a_minus_b = select(a).except_(select(b))
    b_minus_a = select(b).except_(select(a))

    diff_union = union_all(a_minus_b, b_minus_a)

    return _test(statement=diff_union)


def has_at_least_one_not_null_value(column: ColumnElement) -> Compilable:
    """
    Asserts if column has at least one value.

    Example SQL:

    ```sql
    SELECT
        count({{ column_name }}) as filler_column
    FROM
        {{ model }}
    HAVING
        count({{ column_name }}) = 0
    ```

    Example:

    ```python
    has_at_least_one_not_null_value(Health.value)
    ```
    """
    return select(func.count(column, type_=Integer)).having(func.count(column) == 0)


def are_unique_together(columns: Iterable[ColumnElement]) -> Compilable:
    """
    This test confirms that the combination of columns is unique.
    For example, the combination of month and product is unique,
    however neither column is unique in isolation.

    Example:

    ```python
    are_unique_together([HeartRateAgg.year, HeartRateAgg.month])
    ```

    """
    return select(columns).group_by(*columns).having(func.count(type_=Integer) > 1)


def has_the_same_array_length(columns: Iterable[ColumnElement[ARRAY]]) -> Compilable:
    """
    Asserts that all array columns has the same length

    Example SQL:

    ```sql
    SELECT
        arr_col1,
        arr_col2,
        arr_col3
    FROM
        {{ model }}
    WHERE
        ARRAY_LENGTH(arr_col1) != ARRAY_LENGTH(arr_col2)
        OR ARRAY_LENGTH(arr_col1) != ARRAY_LENGTH(arr_col3)
        OR ARRAY_LENGTH(arr_col2) != ARRAY_LENGTH(arr_col3)
    ```

    Example:

    ```python
    assert that([Model.arr_col1, Model.arr_col2, Model.arr_col3], has_the_same_array_length)
    ```
    """
    conditions = (
        func.array_length(a) != func.array_length(b)
        for (a, b) in itertools.combinations(columns, 2)
    )
    return select(columns).where(or_(*conditions))



from typing import Dict, Iterable, List, Tuple

import networkx as nx
from matplotlib import pyplot as plt

from amora.config import settings
from amora.materialization import Task
from amora.models import Column, Model
from amora.utils import list_target_files

CytoscapeElements = List[Dict]


class DependencyDAG(nx.DiGraph):
    def __iter__(self):
        return nx.topological_sort(self)

    @classmethod
    def from_model(cls, model: Model) -> "DependencyDAG":
        """
        Builds the DependencyDAG for a given data model
        """
        dag = cls()
        dag.add_node(model.unique_name())

        def fetch_edges(node: Model):
            for dependency in getattr(node, "__depends_on__", []):
                dag.add_edge(dependency.unique_name(), node.unique_name())
                fetch_edges(dependency)

        fetch_edges(model)
        return dag

    @classmethod
    def from_tasks(cls, tasks: Iterable[Task]) -> "DependencyDAG":
        dag = cls()

        for task in tasks:
            dag.add_node(task.model.unique_name())
            for dependency in getattr(task.model, "__depends_on__", []):
                dag.add_edge(dependency.unique_name(), task.model.unique_name())

        return dag

    @classmethod
    def from_target(cls, models=None) -> "DependencyDAG":
        """
        Builds a DependencyDAG from the files compiled at `settings.AMORA_TARGET_PATH`

        :param models:
        :return:
        """
        model_to_task = {}

        for target_file_path in list_target_files():
            if models and target_file_path.stem not in models:
                continue

            task = Task.for_target(target_file_path)
            model_to_task[task.model.unique_name()] = task

        return cls.from_tasks(tasks=model_to_task.values())

    @classmethod
    def from_columns(cls, columns: List[Tuple[Model, Column]]) -> "DependencyDAG":
        dag = cls()

        for model, column in columns:
            dag.add_node(column)
            for dependency in getattr(model, "__depends_on__", []):
                dependency_columns = dependency.__table__.columns
                if column.key in dependency_columns:
                    dag.add_edge(dependency_columns[column.key], column)

        return dag

    def to_cytoscape_elements(self) -> CytoscapeElements:
        """

        Returns itself as a cytoscape schema compatible representation. E.g:

        For a `A --> B` graph:

        ```python
        [
            {"data": {"id": "A", "label": "A"}},
            {"data": {"id": "B", "label": "B"}},
            {"data": {"source": "A", "target": "B"}},
        ]
        ```
        """

        return [
            *(
                {"group": "nodes", "data": {"id": node, "label": node}}
                for node in self.nodes
            ),
            *(
                {"group": "edges", "data": {"source": source, "target": target}}
                for source, target in self.edges
            ),
        ]

    def draw(self) -> None:
        plt.figure(1, figsize=settings.CLI_MATERIALIZATION_DAG_FIGURE_SIZE)
        nx.draw(
            self,
            with_labels=True,
            font_weight="bold",
            font_size="12",
            linewidths=4,
            node_size=150,
            node_color="white",
            font_color="green",
        )
        plt.show()

    def root(self):
        sorted_elements = list(self)
        if not sorted_elements:
            return None
        return sorted_elements[0]



from datetime import date
from typing import Callable, Set

import pandas as pd
from sqlalchemy.sql import Selectable

from amora.compilation import compile_statement
from amora.protocols import Compilable
from amora.providers.bigquery import run
from amora.storage import cache
from amora.visualization import Table, Visualization, VisualizationConfig

QuestionFunc = Callable[[], Compilable]


class Question:
    """
    A Data Question is a __query__, its __results__, and its __visualization__.

    The __query__ is expressed as an `amora.types.Compilable`, its __results__ is a
    `pandas.DataFrame` and the results __visualization__ as an `amora.views.View`.

    ## Example

    Lets define a new data question:

    ```python
    from examples.models.step_count_by_source import StepCountBySource

    from amora.models import select
    from amora.questions import question


    @question
    def what_are_the_available_data_sources():
        return select(StepCountBySource.source_name).distinct()
    ```

    ### Query
    The query is accessible through the `sql` property:

    ```python
    what_are_the_available_data_sources.sql
    ```

    ```sql
    SELECT DISTINCT `step_count_by_source`.`source_name`
    FROM `amora-data-build-tool.amora`.`step_count_by_source`
    ```

    ### Results

    ```python
    what_are_the_available_data_sources.answer_df()
    ```

    ```
        source_name
    0  Diogo iPhone
    1        Mi Fit
    2        iPhone
    ```

    ### Visualization


    ```python
    what_are_the_available_data_sources.render()
    ```

    |    | source_name   |
    |---:|:--------------|
    |  0 | Diogo iPhone  |
    |  1 | Mi Fit        |
    |  2 | iPhone        |
    """

    def __init__(
        self, question_func: QuestionFunc, view_config: VisualizationConfig = None
    ):
        if isinstance(question_func, Question):
            question_func = question_func.question_func
        if not isinstance(question_func(), Selectable):
            raise ValueError(
                "Valid questions should return a `Selectable`. "
                "Please refer to the 'data questions' documentation."
            )

        self.question_func = question_func
        self.view_config = view_config or Table(title=self.name)

    def __call__(self, *args, **kwargs):
        return self.question_func(*args, **kwargs)

    @property
    def name(self) -> str:
        '''
        Human readable question name. E.g.:

        ```python
        @question
        def what_is_the_most_active_user():
            ...
        ```

        Would result in the parsed function name:
        ```
        >>> what_is_the_most_active_user.name
        "What is the most active user?"
        ```

        But if a question has a docstring, it will be used as the question name:

        ```python
        @question
        def question_2():
            """
            What is the answer to the Ultimate Question of Life, the Universe, and Everything?
            """
            return select(literal(42))
        ```

        ```
        >>> question_2.name
        "What is the answer to the Ultimate Question of Life, the Universe, and Everything?"
        ```
        '''
        if self.question_func.__doc__:
            return self.question_func.__doc__.strip()

        if self.question_func.__name__ == "<lambda>":
            raise NotImplementedError

        question = self.question_func.__name__.replace("_", " ")
        return question.capitalize() + "?"

    @property
    def sql(self) -> str:
        """
        Returns the Question as SQL. E.g:

        ```sql
        SELECT sum(`step_count_by_source`.`value_count`) AS `total`
        FROM `amora-data-build-tool.amora`.`step_count_by_source`
        ```
        """
        stmt = self.question_func()
        return compile_statement(stmt)

    @cache(suffix=lambda self: f"{self.question_func.__name__}.{date.today()}")
    def answer_df(self) -> pd.DataFrame:
        """
        Executes the question against the target database,
        returning a `pandas.DataFrame` as the answer.

        ```python
        what_is_the_current_estimated_walked_distance.answer_df()
        ```

        ```
           total_in_centimeters  total_in_meters  total_in_kilometers   source_name
        0            93030637.0        930306.37            930.30637  Diogo iPhone
        1           112413129.0       1124131.29           1124.13129        Mi Fit
        2            11644600.0        116446.00            116.44600        iPhone
        ```
        """
        result = run(self.question_func())
        return result.rows.to_dataframe(create_bqstorage_client=False)

    @property
    def uid(self) -> str:
        return str(hash(self))

    def render(self) -> Visualization:
        """
        Renders the visual representation of the question's answer
        """
        return Visualization(data=self.answer_df(), config=self.view_config)

    def to_markdown(self) -> str:
        return f"""
            ## {self.name}

            ```sql
            {self.sql}
            ```

            ### Answer

            {self.answer_df().to_markdown()}
        """

    def _repr_markdown_(self):
        return self.to_markdown()

    def __repr__(self):
        return self._repr_markdown_()

    def __eq__(self, other):
        if not isinstance(other, Question):
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        code = self.question_func.__code__
        return hash(code.co_filename + code.co_name)


QUESTIONS: Set[Question] = set()


def question(view_config: VisualizationConfig = None):
    """
    Wraps the function into a `amora.questions.Question`.
    The decorated function must return a `Compilable`

    E.g:
    ```python
    @question()
    def what_are_the_available_data_sources():
        return select(StepCountBySource.source_name).distinct()
    ```

    Optional data visualization configuration can be passed using the `view_config` arg:

    ```python
    from amora.visualization import PieChart


    @question(view_config=PieChart(values="total", names="source_name"))
    def what_is_the_total_step_count_to_date():
        return select(
            func.sum(StepCountBySource.value_sum).label("total"),
            StepCountBySource.source_name,
        ).group_by(StepCountBySource.source_name)
    ```

    """

    def decorator(question_func: QuestionFunc) -> Question:
        q = Question(question_func, view_config=view_config)
        QUESTIONS.add(q)

        return q

    return decorator



import json
from dataclasses import dataclass
from typing import List, Optional

import typer
from jinja2 import Environment, PackageLoader, select_autoescape
from rich.console import Console
from rich.table import Table
from rich.text import Text
from shed import shed

from amora.config import settings
from amora.models import Model, list_models
from amora.providers.bigquery import (
    BIGQUERY_TYPES_TO_PYTHON_TYPES,
    DryRunResult,
    dry_run,
    estimated_query_cost_in_usd,
    estimated_storage_cost_in_usd,
    get_schema,
)

app = typer.Typer(help="List or import Amora Models")


@app.command(name="list")
def models_list(
    format: str = typer.Option(
        "table",
        help="Output format. Options: json,table",
    ),
    with_total_bytes: bool = typer.Option(
        False,
        help="Uses BigQuery query dry run feature "
        "to gather model total bytes information",
    ),
) -> None:
    """
    List the models in your project as a human readable table
    or as a JSON serialized document

    ```shell
    amora models list
    ```
    You can also use the option `--with-total-bytes` to use
    BigQuery query dry run feature to gather model total bytes information

    ```shell
    amora models list --with-total-bytes
    ```

    """

    @dataclass
    class ResultItem:
        model: Model
        dry_run_result: Optional[DryRunResult] = None

        def as_dict(self):
            return {
                "depends_on": self.depends_on,
                "has_source": self.has_source,
                "materialization_type": self.materialization_type,
                "model_name": self.model_name,
                "referenced_tables": self.referenced_tables,
                "total_bytes": self.total_bytes,
                "estimated_query_cost_in_usd": self.estimated_query_cost_in_usd,
                "estimated_storage_cost_in_usd": self.estimated_storage_cost_in_usd,
            }

        @property
        def model_name(self):
            return self.model.__name__

        @property
        def has_source(self):
            return self.model.source() is not None

        @property
        def depends_on(self) -> List[str]:
            return sorted(
                dependency.__name__ for dependency in self.model.dependencies()
            )

        @property
        def estimated_query_cost_in_usd(self) -> Optional[str]:
            if self.dry_run_result:
                cost = estimated_query_cost_in_usd(self.dry_run_result.total_bytes)
                return f"{cost:.{settings.MONEY_DECIMAL_PLACES}f}"
            return None

        @property
        def estimated_storage_cost_in_usd(self) -> Optional[str]:
            if self.dry_run_result:
                cost = estimated_storage_cost_in_usd(self.dry_run_result.total_bytes)
                return f"{cost:.{settings.MONEY_DECIMAL_PLACES}f}"
            return None

        @property
        def total_bytes(self) -> Optional[int]:
            if self.dry_run_result:
                return self.dry_run_result.total_bytes
            return None

        @property
        def referenced_tables(self) -> List[str]:
            if self.dry_run_result:
                return self.dry_run_result.referenced_tables
            return []

        @property
        def materialization_type(self) -> Optional[str]:
            if self.has_source:
                return self.model.__model_config__.materialized.value

            return None

    results = []
    placeholder = "-"

    for model, _model_file_path in list_models():
        if with_total_bytes:
            result_item = ResultItem(model=model, dry_run_result=dry_run(model))
        else:
            result_item = ResultItem(model=model, dry_run_result=None)

        results.append(result_item)

    if format == "table":
        table = Table(
            show_header=True,
            header_style="bold",
            show_lines=True,
            width=settings.CLI_CONSOLE_MAX_WIDTH,
            row_styles=["none", "dim"],
        )

        table.add_column("Model name", style="green bold", no_wrap=True)
        table.add_column("Total bytes", no_wrap=True)
        table.add_column("Estimated query cost", no_wrap=True)
        table.add_column("Estimated storage cost", no_wrap=True)
        table.add_column("Referenced tables")
        table.add_column("Depends on")
        table.add_column("Has source?", no_wrap=True, justify="center")
        table.add_column("Materialization", no_wrap=True)

        for result in results:
            table.add_row(
                result.model_name,
                f"{result.total_bytes or placeholder}",
                result.estimated_query_cost_in_usd or placeholder,
                result.estimated_storage_cost_in_usd or placeholder,
                Text(
                    "\n".join(result.referenced_tables) or placeholder,
                    overflow="fold",
                ),
                Text("\n".join(result.depends_on) or placeholder, overflow="fold"),
                "🟢" if result.has_source else "🔴",
                result.materialization_type or placeholder,
            )

        console = Console(width=settings.CLI_CONSOLE_MAX_WIDTH)
        console.print(table)

    elif format == "json":
        output = {"models": [result.as_dict() for result in results]}
        typer.echo(json.dumps(output))


@app.command(name="import")
def models_import(
    table_reference: str = typer.Option(
        ...,
        "--table-reference",
        help="BigQuery unique table identifier. "
        "E.g.: project-id.dataset-id.table-id",
    ),
    model_file_path: str = typer.Argument(
        None,
        help="Canonical name of python module for the generated AmoraModel. "
        "A good pattern would be to use an unique "
        "and deterministic identifier, like: `project_id.dataset_id.table_id`",
    ),
    overwrite: bool = typer.Option(
        False, help="Overwrite the output file if one already exists"
    ),
):
    """
    Generates a new amora model file from an existing table/view

    ```shell
    amora models import --table-reference my_gcp_project.my_dataset.my_table my_gcp_project/my_dataset/my_table
    ```
    """

    env = Environment(
        loader=PackageLoader("amora"),
        autoescape=select_autoescape(),
        trim_blocks=True,
        lstrip_blocks=True,
    )
    template = env.get_template("new-model.py.jinja2")

    project, dataset, table = table_reference.split(".")
    model_name = "".join(part.title() for part in table.split("_"))

    destination_file_path = settings.models_path.joinpath(
        (model_file_path or model_name.replace(".", "/")) + ".py"
    )

    if destination_file_path.exists() and not overwrite:
        typer.echo(
            f"`{destination_file_path}` already exists. "
            f"Pass `--overwrite` to overwrite file.",
            err=True,
        )
        raise typer.Exit(1)

    sorted_schema = sorted(get_schema(table_reference), key=lambda field: field.name)
    model_source_code = template.render(
        BIGQUERY_TYPES_TO_PYTHON_TYPES=BIGQUERY_TYPES_TO_PYTHON_TYPES,
        dataset=dataset,
        dataset_id=f"{project}.{dataset}",
        model_name=model_name,
        project=project,
        schema=sorted_schema,
        table=table,
    )
    formatted_source_code = shed(model_source_code)

    destination_file_path.parent.mkdir(parents=True, exist_ok=True)
    destination_file_path.write_text(data=formatted_source_code)

    typer.secho(
        f"🎉 Amora Model `{model_name}` (`{table_reference}`) imported!",
        fg=typer.colors.GREEN,
        bold=True,
    )
    typer.secho(f"Current File Path: `{destination_file_path.as_posix()}`")



from dash import dash_table
from dash.development.base_component import Component

from amora.meta_queries import summarize
from amora.models import Model


def component(model: Model) -> Component:
    summary = summarize(model)

    first_column = "column_name"
    return dash_table.DataTable(
        id="model-summary",
        columns=[{"name": first_column, "id": first_column, "selectable": True}]
        + [
            {"name": col, "id": col, "selectable": True}
            for col in sorted(summary.columns.values)
            if col != first_column
        ],
        data=summary.to_dict("records"),
        row_selectable="multi",
        sort_mode="multi",
        sort_by=[{"column_id": first_column, "direction": "asc"}],
        sort_action="native",
        hidden_columns=["is_fv_feature", "is_fv_entity", "is_fv_event_timestamp"],
        style_cell={"overflow": "hidden", "textOverflow": "ellipsis", "maxWidth": 0},
        style_data_conditional=[
            {
                "if": {
                    "filter_query": "{is_fv_entity} >= 1",
                    "column_id": "column_name",
                },
                "backgroundColor": "#ddccdd",
            },
            {
                "if": {
                    "filter_query": "{is_fv_feature} >= 1",
                    "column_id": "column_name",
                },
                "backgroundColor": "#ddddcc",
            },
            {
                "if": {
                    "filter_query": "{is_fv_event_timestamp} >= 1",
                    "column_id": "column_name",
                },
                "backgroundColor": "#ccdddd",
            },
        ],
        export_format="csv",
        export_headers="display",
    )



from typing import Iterable, Tuple

import dash
import dash_bootstrap_components as dbc
from dash import html
from dash.development.base_component import Component

from amora.config import settings as base_settings
from amora.dash.config import settings as dash_settings
from amora.feature_store.config import settings as feature_store_settings

dash.register_page(__name__, fa_icon="fa-gear", location="sidebar")


def get_environment_data() -> Iterable[Tuple[str, str]]:
    def gen_data():
        for settings in (base_settings, feature_store_settings, dash_settings):
            for k, v in settings.dict().items():
                yield f"{settings.Config.env_prefix}{k}", str(v)

    return sorted(gen_data())


def layout() -> Component:
    return dbc.Table(
        [
            html.Thead(html.Tr([html.Th("Key"), html.Th("Value")])),
            html.Tbody(
                [
                    html.Tr([html.Td(key), html.Td(value)])
                    for key, value in get_environment_data()
                ]
            ),
        ],
        id="environment-table",
    )



from typing import Iterable, Tuple

import dash_bootstrap_components as dbc
import dash_cytoscape
from dash import Input, Output, callback, dcc
from dash.development.base_component import Component

from amora.dag import CytoscapeElements, DependencyDAG
from amora.feature_store.protocols import FeatureViewSourceProtocol
from amora.models import MaterializationTypes, Model, amora_model_for_name


def _style_elements(elements: CytoscapeElements) -> CytoscapeElements:
    def border_color_for_model(model: Model) -> str:
        if isinstance(model, FeatureViewSourceProtocol):
            return "green"
        return "grey"

    def background_color_for_model(model: Model) -> str:
        return {
            MaterializationTypes.table: "black",
            MaterializationTypes.view: "grey",
            MaterializationTypes.ephemeral: "white",
        }[model.__model_config__.materialized]

    def node_models() -> Iterable[Tuple[str, Model]]:
        for elem in elements:
            if elem["group"] != "nodes":
                continue

            model_name = elem["data"]["id"]
            model = amora_model_for_name(model_name)

            yield model_name, model

    return [
        *(
            {
                "data": {
                    "id": node,
                    "label": node,
                },
                "style": {
                    "border-color": border_color_for_model(model),
                    "border-width": 3,
                    "background-color": background_color_for_model(model),
                },
            }
            for node, model in node_models()
        ),
        *(edge for edge in filter(lambda e: e["group"] == "edges", elements)),
    ]


def component(dag: DependencyDAG, height: str = "400px") -> Component:
    elements = dag.to_cytoscape_elements()
    elements = _style_elements(elements)

    return dbc.Row(
        className="cy-container",
        children=[
            dash_cytoscape.Cytoscape(
                id="cytoscape-layout",
                elements=elements,
                layout={
                    "name": "breadthfirst",
                    "refresh": 20,
                    "fit": True,
                    "padding": 30,
                    "randomize": False,
                },
                style={"width": "100%", "height": height},
                stylesheet=[
                    {"selector": "node", "style": {"label": "data(label)"}},
                    {
                        "selector": "edge",
                        "style": {
                            "curve-style": "bezier",
                            "target-arrow-shape": "triangle",
                        },
                    },
                ],
                responsive=True,
            ),
            dbc.Row(id="cytoscape-output"),
        ],
    )


@callback(
    Output("cytoscape-output", "children"),
    Input("cytoscape-layout", "tapNodeData"),
    prevent_initial_call=True,
)
def redirect_to_model(data):
    unique_identifier = data["id"]
    return dcc.Location(pathname=f"/models/{unique_identifier}", id=unique_identifier)



from typing import Union

import pytest
from _pytest.config import ExitCode
from _pytest.main import Session
from dash.testing.composite import DashComposite
from rich.console import Console
from rich.table import Table

from amora.config import settings
from amora.dash.app import dash_app
from amora.tests.audit import AuditLog


def pytest_sessionstart():
    print(
        """
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A amora adoça mais na boca de quem namora.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Mantido com <3 por `RexData@Stone`
"""
    )


def pytest_sessionfinish(session: Session, exitstatus: Union[int, ExitCode]) -> None:
    log_rows = AuditLog.get_all(test_run_id=settings.TEST_RUN_ID)

    table = Table(
        show_header=True,
        header_style="bold",
        show_lines=True,
        width=settings.CLI_CONSOLE_MAX_WIDTH,
        row_styles=["none", "dim"],
    )
    table.add_column("🧪 Test node id")
    table.add_column("🔎 Bytes billed")
    table.add_column("💰 Estimated cost (USD)")
    for audit_log in log_rows:
        table.add_row(
            audit_log.test_node_id,
            str(audit_log.bytes_billed),
            str(audit_log.estimated_cost_in_usd),
        )

    console = Console(width=settings.CLI_CONSOLE_MAX_WIDTH)
    console.print(table, new_line_start=True)


@pytest.fixture
def amora_dash(dash_duo: DashComposite) -> DashComposite:
    dash_duo.start_server(dash_app)
    return dash_duo



from dash import dcc, html
from dash.development.base_component import Component

from amora.dashboards import DateFilter


def layout(filter: DateFilter) -> Component:
    return html.Div(
        [
            filter.title,
            dcc.DatePickerSingle(
                id=filter.id,
                min_date_allowed=filter.min_selectable_date,
                max_date_allowed=filter.max_selectable_date,
                initial_visible_month=filter.default,
                date=filter.default,
            ),
        ]
    )



import dash
import dash_bootstrap_components as dbc
from dash import Input, Output, dcc, html
from dash.development.base_component import Component
from dash_extensions import Lottie

from amora.dash.components import question_details
from amora.dash.components.animation import Lotties
from amora.dash.components.filters import filter
from amora.dashboards import DASHBOARDS, Dashboard, list_dashboards

dash.register_page(
    __name__,
    fa_icon="fa-chart-line",
    location="sidebar",
    path="/dashboards",
    path_template="/dashboards/<dashboard_id>",
)

list_dashboards()


def render(dashboard: Dashboard) -> Component:
    questions = [
        dbc.Row(
            children=[
                dbc.Col(question_details.component(question_col))
                for question_col in row
            ]
        )
        for row in dashboard.questions
    ]
    filters = [filter.layout(f) for f in dashboard.filters]

    return html.Div([html.Div(filters), html.Div(questions)])


def dashboards_dropdown() -> Component:
    options = [
        {"label": dashboard.name, "value": dashboard.uid}
        for dashboard in DASHBOARDS.values()
    ]
    return dcc.Dropdown(
        options=options,
        id="dashboard-select-dropdown",
        value=None,
        placeholder="Select a dashboard",
    )


def dashboards_selector() -> Component:
    return html.Div(
        [
            html.H1("🧑‍🔬 Select a dashboard and start exploring"),
            dashboards_dropdown(),
            Lottie(
                options=dict(
                    loop=True,
                    autoplay=True,
                    rendererSettings=dict(preserveAspectRatio="xMidYMid slice"),
                ),
                width="70%",
                url=Lotties.green_bi_dashboard,
            ),
        ],
        id="dashboard-content",
        style={"min-height": "600px"},
    )


def layout(dashboard_id: str = None) -> Component:
    if not dashboard_id:
        return dashboards_selector()

    dashboard = DASHBOARDS[dashboard_id]
    return html.Div([html.H1(dashboard.name), render(dashboard)])


@dash.callback(
    Output("dashboard-content", "children"),
    Input("dashboard-select-dropdown", "value"),
    prevent_initial_call=True,
)
def update_dashboard_details(value: str) -> Component:
    return render(dashboard=DASHBOARDS[value])



import dash
from dash import html

dash.register_page(__name__)


def layout():
    return html.Div(
        [
            html.H1("404: Not found", className="text-danger", id="page-not-found"),
            html.Hr(),
            html.P("The pathname was not recognised..."),
        ]
    )



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from google.cloud.bigquery import Client, QueryJobConfig, Table

from amora.models import MaterializationTypes, Model, amora_model_for_target_path


@dataclass
class Task:
    sql_stmt: str
    model: Model
    target_file_path: Path

    @classmethod
    def for_target(cls, target_file_path: Path) -> "Task":
        return cls(
            sql_stmt=target_file_path.read_text(),
            model=amora_model_for_target_path(target_file_path),
            target_file_path=target_file_path,
        )

    def __repr__(self):
        return f"{self.model.__name__} -> {self.sql_stmt}"


def materialize(sql: str, model: Model) -> Optional[Table]:
    config = model.__model_config__
    materialization = config.materialized

    if materialization == MaterializationTypes.ephemeral:
        return None

    client = Client()

    if materialization == MaterializationTypes.view:
        table_name = model.unique_name()

        view = Table(table_name)
        view.description = config.description
        view.labels = config.labels_dict
        view.view_query = sql

        client.delete_table(table_name, not_found_ok=True)

        return client.create_table(view)

    if materialization == MaterializationTypes.table:
        table_name = model.unique_name()
        client.delete_table(table_name, not_found_ok=True)
        query_job = client.query(
            sql,
            job_config=QueryJobConfig(
                destination=model.unique_name(),
                write_disposition="WRITE_TRUNCATE",
            ),
        )

        query_job.result()

        table = client.get_table(model.unique_name())
        table.description = config.description
        table.labels = config.labels_dict

        if config.cluster_by:
            table.clustering_fields = config.cluster_by

        return client.update_table(
            table, ["description", "labels", "clustering_fields"]
        )

    raise ValueError(
        f"Invalid model materialization configuration. "
        f"Valid types are: `{', '.join((m.name for m in MaterializationTypes))}`. "
        f"Got: `{materialization}`"
    )



import inspect
import re
from dataclasses import dataclass, field
from enum import Enum, auto
from importlib.util import module_from_spec, spec_from_file_location
from inspect import getfile
from pathlib import Path
from types import ModuleType
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

from sqlalchemy import Column, MetaData, Table, select
from sqlalchemy.orm import declared_attr
from sqlalchemy.sql import ColumnElement
from sqlmodel import Field, Session, SQLModel, create_engine

from amora.config import settings
from amora.logger import logger
from amora.protocols import Compilable, CompilableProtocol
from amora.utils import list_files, model_path_for_target_path

select = select
Column = Column
ColumnElement = ColumnElement
Columns = Iterable[ColumnElement]
Field = Field
Model = Type["AmoraModel"]
MetaData = MetaData
Models = Iterable[Model]
Session = Session
create_engine = create_engine

LabelKey = str
LabelValue = str
LabelKeys = Iterable[LabelKey]


class Label(NamedTuple):
    """
    A label is a (key, value) pair that can also be represented as a "key:value" string. E.g.:

    ```python
    Label("freshness", "daily")
    ```
    """

    key: LabelKey
    value: LabelValue

    def __str__(self):
        return f"{self.key}:{self.value}"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if not isinstance(other, str):
            return self.key == other.key and self.value == other.value

        return self == Label.from_str(other)

    @classmethod
    def from_str(cls, label: str):
        """
        >>> Label.from_str("domain:health")
        Label(key="domain", value="health")
        """

        return cls(*label.split(":"))


LabelRepr = Union[Label, str]
Labels = Set[Label]


@dataclass
class PartitionConfig:
    field: str
    data_type: str = "date"
    granularity: str = "day"
    range: Optional[Dict[str, Any]] = None


class AutoName(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name


class MaterializationTypes(AutoName):
    ephemeral = auto()
    view = auto()
    table = auto()


@dataclass
class ModelConfig:
    """
    Model configuration metadata

    Attributes:
        cluster_by (List[str]): BigQuery tables can be [clustered](https://cloud.google.com/bigquery/docs/clustered-tables) to colocate related data. Expects a list of columns, as strings.
        description (Optional[str]): A string description of the model, used for documentation
        labels (Labels): Labels that can be used for data catalog and resource selection
        materialized (amora.models.MaterializationTypes): The materialization configuration: `view`, `table`, `ephemeral`. Default: `view`
        partition_by (Optional[PartitionConfig]): BigQuery supports the use of a [partition by](https://cloud.google.com/bigquery/docs/partitioned-tables) clause to easily partition a table by a column or expression. This option can help decrease latency and cost when querying large tables.
    """

    description: str = "Undocumented! Generated by Amora Data Build Tool 💚"
    materialized: MaterializationTypes = MaterializationTypes.view
    partition_by: Optional[PartitionConfig] = None
    cluster_by: List[str] = field(default_factory=list)
    labels: Labels = field(default_factory=set)

    @property
    def labels_dict(self) -> Dict[str, str]:
        return {label.key: label.value for label in self.labels}


metadata = MetaData(schema=f"{settings.TARGET_PROJECT}.{settings.TARGET_SCHEMA}")


class AmoraModel(SQLModel):
    """
    Attributes:
        __depends_on__ (Models): A list of Amora Models that the current model depends on
        __model_config__ (ModelConfig): Model configuration metadata
        __table__ (Table): SQLAlchemy table object
        __table_args__ (Dict[str, Any]): SQLAlchemy table arguments
    """

    __depends_on__: Models = []
    __model_config__ = ModelConfig(materialized=MaterializationTypes.view)
    __table__: Table
    __table_args__: Dict[str, Any] = {"extend_existing": True}
    metadata = metadata

    @declared_attr  # type: ignore
    def __tablename__(cls) -> str:  # type: ignore
        """
        By default, `__tablename__` is the `snake_case` class name.

        ```python
        class MyModel(AmoraModel):
            ...


        assert MyModel.__tablename__ == "my_model"
        ```
        """
        return re.sub(r"(?<!^)(?=[A-Z])", "_", cls.__name__).lower()

    @classmethod
    def dependencies(cls) -> Models:
        source = cls.source()
        if source is None:
            return []

        # todo: Remover necessidade de __depends_on__ inspecionando a query e chegando ao modelo de origem
        # tables: List[Table] = source.froms

        return cls.__depends_on__

    @classmethod
    def source(cls) -> Optional[Compilable]:
        """
        Called when `amora compile` is executed, Amora will build this model
        in your data warehouse by wrapping it in a `create view as` or `create table as` statement.

        Return `None` for defining models for tables/views that already exist on the data warehouse
        and shouldn't be managed by Amora.

        Return a `Compilable`, which is a sqlalchemy select statement, in order to compile the model with the given statement
        :return:
        """
        return None

    @classmethod
    def target_path(cls, model_file_path: Union[str, Path]) -> Path:
        # {settings.dbt_models_path}/a_model/a_model.py -> a_model/a_model.py
        strip_path = settings.models_path.as_posix()
        relative_model_path = str(model_file_path).split(strip_path)[1][1:]
        # a_model/a_model.py -> ~/project/amora/target/a_model/a_model.sql
        target_file_path = settings.target_path.joinpath(
            relative_model_path.replace(".py", ".sql")
        )

        return target_file_path

    @classmethod
    def model_file_path(cls) -> Path:
        return Path(getfile(cls))

    @classmethod
    def unique_name(cls) -> str:
        return str(cls.__table__)


def _is_amora_model(candidate: ModuleType) -> bool:
    return (
        isinstance(candidate, CompilableProtocol)
        and inspect.isclass(candidate)
        and issubclass(candidate, AmoraModel)
        and hasattr(candidate, "__table__")
    )


def amora_model_for_path(path: Path) -> Model:
    spec = spec_from_file_location(".".join(["amoramodel", path.stem]), path)
    if spec is None:
        raise ValueError(f"Invalid path `{path}`. Not a valid Python file.")

    module = module_from_spec(spec)

    if spec.loader is None:
        raise ValueError(f"Invalid AmoraModel path `{path}`. Unable to load module.")

    try:
        spec.loader.exec_module(module)
    except ImportError as e:
        raise ValueError(
            f"Invalid AmoraModel path `{path}`. Unable to load module."
        ) from e

    compilables = inspect.getmembers(
        module,
        _is_amora_model,
    )

    for _name, class_ in compilables:
        try:
            # fixme: Quando carregamos o código em inspect, não existe um arquivo associado,
            #  ou seja, ao iterar sobre as classes de um arquivo, a classe que retornar um TypeError,
            #  é uma classe definida no arquivo
            class_.model_file_path()
        except TypeError:
            return class_

    raise ValueError(f"Invalid path `{path}`")


def amora_model_for_target_path(path: Path) -> Model:
    model_path = model_path_for_target_path(path)
    return amora_model_for_path(model_path)


def model_path_for_model(model: Model) -> Path:
    """
    Returns the filepath where the model is defined.
    """
    for m, path in list_models():
        if m.unique_name() == model.unique_name():
            return path
    raise FileNotFoundError("Model file not found in the project")


def amora_model_for_name(model_name: str) -> Model:
    for model, _path in list_models():
        if model.unique_name() == model_name:
            return model
    raise ValueError(f"{model_name} not found on models list")


def list_models(
    path: Path = settings.models_path,
) -> Iterable[Tuple[Model, Path]]:
    for model_file_path in list_files(path, suffix=".py"):
        if model_file_path.stem.startswith("_"):
            continue
        try:
            yield amora_model_for_path(model_file_path), model_file_path
        except ValueError:
            logger.exception(
                "Unable to load amora model for path",
                extra={"model_file_path": model_file_path},
            )
            continue


def select_models_with_labels(labels: Labels) -> Iterable[Tuple[Model, Path]]:
    def matches_labels(item: Tuple[Model, Path]) -> bool:
        model, _model_path = item
        return match_labels(model, labels)

    return filter(matches_labels, list_models())


def select_models_with_label_keys(
    label_keys: LabelKeys,
) -> Iterable[Tuple[Model, Path]]:
    keys = set(label_keys)

    def matches_labels(item: Tuple[Model, Path]) -> bool:
        model, _model_path = item
        return match_label_keys(model, keys)

    return filter(matches_labels, list_models())


def match_label_keys(model: Model, label_keys: Iterable[LabelKey]) -> bool:
    for key in label_keys:
        for label in model.__model_config__.labels:
            if label.key == key:
                return True
    return False


def match_labels(model: Model, labels: Labels) -> bool:
    for label in labels:
        if label in model.__model_config__.labels:
            return True
    return False



from datetime import date
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from amora.config import settings
from amora.logger import logger
from amora.questions import Question
from amora.utils import list_files


class Filter(BaseModel):
    type: str
    id: str
    default: Any
    title: str


class DateFilter(Filter):
    type = "date"
    default: date = date.today()
    python_type = date
    min_selectable_date: Optional[date] = None
    max_selectable_date: Optional[date] = None


class AcceptedValuesFilter(Filter):
    type = "accepted_values"
    values: List[str]
    default: Optional[str] = None

    # todo: validate that "self.default in self.values"


DashboardUid = str


class Dashboard(BaseModel):
    """
    A set of one or more questions, organized and arranged into one or more rows,
    that provide an at-a-glance view of related information. Dashboards can optionally
    contain `filters`, that act upon `questions`, refining it's answers.

    A new dashboard should be defined as a Python module with a `dashboard: Dashboard`
    attribute. E.g:

    ```python
    # $AMORA_PROJECT_PATH/dashboards/steps.py

    from amora.dashboards import AcceptedValuesFilter, Dashboard, DateFilter
    from examples.amora_project.models.step_count_by_source import (
        how_many_data_points_where_acquired,
        what_are_the_available_data_sources,
        what_are_the_values_observed_on_the_iphone,
        what_is_the_current_estimated_walked_distance,
        what_is_the_latest_data_point,
        what_is_the_total_step_count_to_date,
    )

    dashboard = Dashboard(
        uid="1",
        name="Health :: Step Analysis",
        questions=[
            [
                what_is_the_latest_data_point,
                what_is_the_current_estimated_walked_distance,
                how_many_data_points_where_acquired,
            ],
            [
                what_is_the_total_step_count_to_date,
                what_are_the_values_observed_on_the_iphone,
                what_are_the_available_data_sources,
            ],
        ],
        filters=[
            DateFilter(
                default="2021-01-01", title="data de início", id="start-date-filter"
            ),
            DateFilter(default="2023-01-01", title="data fim", id="end-date-filter"),
            AcceptedValuesFilter(
                default="iPhone",
                values=["Diogo's iPhone", "iPhone"],
                title="Source device",
                id="source-device-filter",
            ),
        ],
    )
    ```
    """

    uid: DashboardUid
    name: str
    questions: List[List[Question]]
    filters: List[Filter]

    class Config:
        arbitrary_types_allowed = True


DASHBOARDS: Dict[DashboardUid, Dashboard] = dict()


def dashboard_for_path(path: Path) -> Dashboard:
    spec = spec_from_file_location(".".join(["amoradashboard", path.stem]), path)
    if spec is None:
        raise ValueError(f"Invalid path `{path}`. Not a valid Python file.")

    module = module_from_spec(spec)

    if spec.loader is None:
        raise ValueError(f"Invalid Dashboard path `{path}`. Unable to load module.")

    try:
        spec.loader.exec_module(module)  # type: ignore
    except Exception as e:
        raise ValueError(
            f"Invalid Dashboard path `{path}`. Unable to execute module."
        ) from e

    dashboard = getattr(module, "dashboard")
    assert isinstance(dashboard, Dashboard)
    return dashboard


def list_dashboards(
    path: Path = settings.dashboards_path,
) -> Dict[DashboardUid, Dashboard]:
    """
    Searches for python files with
    Args:
        path: The path to search for Dashboard definitions on files

    Returns:
        A dict of all available dashboards
    """
    for file_path in list_files(path, suffix=".py"):
        if file_path.stem.startswith("_"):
            continue

        try:
            dashboard = dashboard_for_path(file_path)
        except ValueError:
            logger.exception("Unable to load dashboard")
        else:
            DASHBOARDS[dashboard.uid] = dashboard

    return DASHBOARDS



from datetime import datetime
from typing import Optional

import typer

from amora.cli.shared_options import models_option
from amora.cli.type_specs import Models

app = typer.Typer(help="Easily productionize new features from Amora Models")


@app.command(name="plan")
def feature_store_plan():
    """
    Dry-run registering objects to the Feature Registry

    The plan method dry-runs registering one or more definitions (e.g.: Entity, Feature View)
    and produces a list of all the changes that would be introduced in the Feature Registry
    by an `amora feature-store apply` execution.

    The changes computed by the `plan` command are informational, and are not actually applied to the registry.
    """
    from amora.feature_store import fs
    from amora.feature_store.registry import get_repo_contents

    registry_diff, infra_diff, _infra = fs._plan(
        desired_repo_contents=get_repo_contents()
    )

    typer.echo("Amora: Feature Store :: Registry diff")
    typer.echo(registry_diff.to_string())

    typer.echo("Amora: Feature Store :: Infrastructure diff")
    typer.echo(infra_diff.to_string())


@app.command(name="apply")
def feature_store_apply():
    """
    1. Scans Python files in your amora project and find all models defined as
    feature views.

    2. Validate your feature definitions

    3. Sync the metadata about feature store objects to the feature registry.
    If a registry does not exist, then it will be instantiated.
    The standard registry is a simple protobuf binary file
    that is stored on disk (locally or in an object store).

    4. Create all necessary feature store infrastructure.
    The exact infrastructure that is deployed or configured depends
    on the provider configuration. For example, setting local as
    your provider will result in a sqlite online store being created.
    """
    from feast.repo_operations import apply_total_with_repo_instance

    from amora.feature_store import fs
    from amora.feature_store.registry import get_repo_contents

    apply_total_with_repo_instance(
        store=fs,
        project=fs.project,
        registry=fs.registry,
        repo=get_repo_contents(),
        skip_source_validation=False,
    )


@app.command(name="materialize")
def feature_store_materialize(
    start_ts: str = typer.Argument(
        None,
        help="Start timestamp on ISO 8601 format. E.g.: '2022-01-01T01:00:00'",
    ),
    end_ts: str = typer.Argument(
        None,
        help="End timestamp on ISO 8601 format. E.g.: '2022-01-02T01:00:00'",
    ),
    models: Optional[Models] = models_option,
):
    """
    Run a (non-incremental) materialization job to ingest data into the online
    store. All data between `start_ts` and `end_ts` will be read from the offline
    store and written into the online store. If you don't specify feature view
    names using `--models`, all registered Feature Views will be materialized.
    """
    from amora.feature_store import fs
    from amora.feature_store.registry import get_repo_contents

    repo_contents = get_repo_contents()

    if models:
        views_to_materialize = [
            fv.name for fv in repo_contents.feature_views if fv.name in models
        ]
    else:
        views_to_materialize = [fv.name for fv in repo_contents.feature_views]

    fs.materialize(
        feature_views=views_to_materialize,
        start_date=datetime.fromisoformat(start_ts),
        end_date=datetime.fromisoformat(end_ts),
    )


@app.command(name="materialize-incremental")
def feature_store_materialize_incremental(
    end_ts: Optional[str] = typer.Argument(
        None,
        help="End timestamp on ISO 8601 format. E.g.: '2022-01-02T01:00:00'. If a date isn't provided, `datetime.utcnow` is used",
    ),
    models: Optional[Models] = models_option,
):
    """
    Load data from feature views into the online store, beginning from either the previous `materialize`
    or `materialize-incremental` end date, or the beginning of time.

    """
    from amora.feature_store import fs
    from amora.feature_store.registry import get_repo_contents

    repo_contents = get_repo_contents()

    if models:
        views_to_materialize = [
            fv.name for fv in repo_contents.feature_views if fv.name in models
        ]
    else:
        views_to_materialize = [fv.name for fv in repo_contents.feature_views]

    if end_ts is not None:
        end_date = datetime.fromisoformat(end_ts)
    else:
        end_date = datetime.utcnow()

    fs.materialize_incremental(
        feature_views=views_to_materialize,
        end_date=end_date,
    )


@app.command(name="serve")
def feature_store_serve():
    """
    Starts the feature server HTTP app.

    Routes:

        - `POST /get-online-features`

        `curl -XPOST -H "Content-type: application/json" -d '{"features": ["step_count_by_source:value_avg", "step_count_by_source:value_sum", "step_count_by_source:value_count"], "entities": {"source_name": ["Mi Fit", "Diogo iPhone", "An invalid source"]}}' 'http://localhost:8666/get-online-features'`

        ```json
        {
          "metadata": {
            "feature_names": [
              "source_name",
              "value_count",
              "value_sum",
              "value_avg"
            ]
          },
          "results": [
            {
              "values": [
                "Mi Fit",
                6.0,
                809.0,
                134.8333282470703
              ],
              "statuses": [
                "PRESENT",
                "PRESENT",
                "PRESENT",
                "PRESENT"
              ],
              "event_timestamps": [
                "1970-01-01T00:00:00Z",
                "2021-07-23T02:00:00Z",
                "2021-07-23T02:00:00Z",
                "2021-07-23T02:00:00Z"
              ]
            },
            {
              "values": [
                "Diogo iPhone",
                2.0,
                17.0,
                8.5
              ],
              "statuses": [
                "PRESENT",
                "PRESENT",
                "PRESENT",
                "PRESENT"
              ],
              "event_timestamps": [
                "1970-01-01T00:00:00Z",
                "2021-07-23T02:00:00Z",
                "2021-07-23T02:00:00Z",
                "2021-07-23T02:00:00Z"
              ]
            },
            {
              "values": [
                "An invalid source",
                null,
                null,
                null
              ],
              "statuses": [
                "PRESENT",
                "NOT_FOUND",
                "NOT_FOUND",
                "NOT_FOUND"
              ],
              "event_timestamps": [
                "1970-01-01T00:00:00Z",
                "2021-07-23T02:00:00Z",
                "2021-07-23T02:00:00Z",
                "2021-07-23T02:00:00Z"
              ]
            }
          ]
        }
        ```

        More on: https://docs.feast.dev/v/v0.9-branch/user-guide/getting-online-features

        - `GET /list-feature-views`. E.g.:

        `curl http://localhost:8666/list-feature-views | jq`

        ```json
        [
            {
                "name": "step_count_by_source",
                "features": [
                    "step_count_by_source:value_avg",
                    "step_count_by_source:value_sum",
                    "step_count_by_source:value_count"
                ],
                "entities": [
                    "source_name"
                ]
            }
        ]
        ```
    """
    import uvicorn
    from feast.feature_server import get_app
    from prometheus_fastapi_instrumentator import Instrumentator

    from amora.feature_store import fs
    from amora.feature_store.config import settings

    app = get_app(store=fs)

    @app.get("/list-feature-views")
    def list_feature_views():
        fvs = fs.list_feature_views()
        return [
            {
                "name": fv.name,
                "features": [f"{fv.name}:{feature.name}" for feature in fv.features],
                "entities": fv.entities,
            }
            for fv in fvs
        ]

    Instrumentator().instrument(app).expose(app)

    uvicorn.run(
        app,
        host=settings.HTTP_SERVER_HOST,
        port=settings.HTTP_SERVER_PORT,
        access_log=settings.HTTP_ACCESS_LOG_ENABLED,
    )



from feast import FeatureStore, RepoConfig

from amora.feature_store.config import settings
from amora.feature_store.logging import patch_tqdm
from amora.feature_store.usage_tracking import patch_usage

patch_usage()
patch_tqdm()

repo_config = RepoConfig(
    registry=settings.REGISTRY,
    project="amora",
    provider=settings.PROVIDER,
    online_store={
        "type": settings.ONLINE_STORE_TYPE,
        **settings.ONLINE_STORE_CONFIG,
    },
    offline_store={
        "type": settings.OFFLINE_STORE_TYPE,
        **settings.OFFLINE_STORE_CONFIG,
    },
    repo_path=settings.REPO_PATH,
)

fs = FeatureStore(config=repo_config)



from typing import Dict, Iterable, List, Tuple

from feast import Entity, FeatureService, FeatureView
from feast.repo_contents import RepoContents
from sqlalchemy.orm import InstrumentedAttribute

from amora.feature_store.feature_view import name_for_model
from amora.feature_store.type_mapping import type_for_column
from amora.models import Model, list_models

FEATURE_REGISTRY: Dict[str, Tuple[FeatureView, FeatureService, Model]] = {}


def get_entities() -> Iterable[Entity]:
    for fv, _service, model in FEATURE_REGISTRY.values():
        for entity_name in fv.entities:
            entity_column: InstrumentedAttribute = getattr(model, entity_name)

            yield Entity(
                name=entity_name,
                value_type=type_for_column(entity_column),
                description=entity_column.comment,
            )


def get_feature_views() -> List[FeatureView]:
    return [fv for (fv, _service, _model) in FEATURE_REGISTRY.values()]


def get_feature_service(model: Model) -> FeatureService:
    (_fv, service, _model) = FEATURE_REGISTRY[name_for_model(model)]
    return service


def get_feature_services() -> List[FeatureService]:
    return [service for (_fv, service, _model) in FEATURE_REGISTRY.values()]


def get_repo_contents() -> RepoContents:
    # fixme: making sure that we've collected all Feature Views
    _models = list(list_models())

    return RepoContents(
        feature_views=set(get_feature_views()),
        entities=set(get_entities()),
        feature_services=set(get_feature_services()),
        on_demand_feature_views=set(),
        request_feature_views=set(),
    )



import os

from amora.feature_store.config import settings


def patch_usage() -> None:
    """
    By default, feast tracks user activity. We believe that should be an opt-in feature, instead of opt-out.

    Reference: https://docs.feast.dev/reference/usage
    """
    from feast import usage
    from feast.constants import FEAST_USAGE

    os.environ[FEAST_USAGE] = str(settings.USAGE_TRACKING_ENABLED)

    usage._is_enabled = settings.USAGE_TRACKING_ENABLED
    usage.USAGE_ENDPOINT = settings.USAGE_ENDPOINT



import decimal
from dataclasses import dataclass
from datetime import date, datetime, time
from enum import Enum
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import pandas as pd
import sqlalchemy
from google.api_core.client_info import ClientInfo
from google.cloud.bigquery import (
    Client,
    QueryJobConfig,
    SchemaField,
    Table,
    TableReference,
)
from google.cloud.bigquery.table import RowIterator, _EmptyRowIterator
from pydantic.fields import SHAPE_LIST
from sqlalchemy import func, literal, literal_column, tablesample
from sqlalchemy.sql import coercions, expression, operators, roles, sqltypes
from sqlalchemy.sql.selectable import CTE
from sqlalchemy.sql.sqltypes import ARRAY
from sqlalchemy_bigquery import STRUCT
from sqlalchemy_bigquery.base import BQArray, BQBinary, unnest
from sqlmodel.main import get_sqlachemy_type
from sqlmodel.sql import sqltypes as sqlmodel_types

from amora import logger
from amora.compilation import compile_statement
from amora.config import settings
from amora.contracts import BaseResult
from amora.models import (
    AmoraModel,
    Column,
    ColumnElement,
    MaterializationTypes,
    Model,
    select,
)
from amora.protocols import Compilable
from amora.storage import cache
from amora.version import VERSION

Schema = List[SchemaField]
BQTable = Union[Table, TableReference, str]

BIGQUERY_TYPES_TO_PYTHON_TYPES = {
    "ARRAY": list,
    "BIGNUMERIC": decimal.Decimal,  # todo: test coverage
    "NUMERIC": decimal.Decimal,  # todo: test coverage
    "BOOL": bool,
    "BOOLEAN": bool,
    "BYTES": bytes,
    "DATE": date,
    "DATETIME": datetime,
    "FLOAT64": float,
    "FLOAT": float,
    "INT64": int,
    "INTEGER": int,
    "JSON": dict,
    "STRING": str,
    "TIME": time,
    "TIMESTAMP": datetime,
    "RECORD": dict,
    "STRUCT": dict,
}

BIGQUERY_TYPES_TO_SQLALCHEMY_TYPES = {
    "ARRAY": BQArray,
    "BIGNUMERIC": sqltypes.Numeric,
    "BOOL": sqltypes.Boolean,
    "BOOLEAN": sqltypes.Boolean,
    "BYTES": BQBinary,
    "DATE": sqltypes.Date,
    "DATETIME": sqltypes.DateTime,
    "FLOAT": sqltypes.Float,
    "FLOAT64": sqltypes.Float,
    "INT64": sqltypes.Integer,
    "INTEGER": sqltypes.Integer,
    "JSON": sqltypes.JSON,
    "NUMERIC": sqltypes.Numeric,
    "RECORD": STRUCT,
    "STRING": sqltypes.String,
    "STRUCT": STRUCT,
    "TIME": sqltypes.Time,
    "TIMESTAMP": sqltypes.TIMESTAMP,
}


SQLALCHEMY_TYPES_TO_BIGQUERY_TYPES = {
    BQBinary: "BYTES",
    STRUCT: "RECORD",
    sqlmodel_types.AutoString: "STRING",
    sqltypes.Boolean: "BOOLEAN",
    sqltypes.Date: "DATE",
    sqltypes.DateTime: "DATETIME",
    sqltypes.Float: "FLOAT",
    sqltypes.Integer: "INTEGER",
    sqltypes.JSON: "JSON",
    sqltypes.Numeric: "NUMERIC",
    sqltypes.String: "STRING",
    sqltypes.TIMESTAMP: "TIMESTAMP",
    sqltypes.Time: "TIME",
}


class TimePart(Enum):
    """
    Since BigQuery represents time parts as an unquoted value,
    we need to use `literal_column`

    More on: https://cloud.google.com/bigquery/docs/reference/standard-sql/time_functions#time_trunc
    """

    MICROSECOND = literal_column("MICROSECOND")
    MILLISECOND = literal_column("MILLISECOND")
    SECOND = literal_column("SECOND")
    MINUTE = literal_column("MINUTE")
    HOUR = literal_column("HOUR")


@dataclass
class DryRunResult(BaseResult):
    model: Model
    schema: Schema

    @property
    def estimated_cost(self):
        return estimated_query_cost_in_usd(self.total_bytes)


@dataclass
class RunResult(BaseResult):
    rows: Union[RowIterator, _EmptyRowIterator]
    execution_time_in_ms: int
    to_dataframe: Callable[..., pd.DataFrame]
    schema: Optional[Schema] = None

    @property
    def estimated_cost(self):
        return estimated_query_cost_in_usd(self.total_bytes)


_client = None


def get_client() -> Client:
    global _client
    if _client is None:
        _client = Client(
            client_info=ClientInfo(
                client_library_version=VERSION,
                user_agent=f"amora-data-build-tool/{VERSION}",
            )
        )
    return _client


def get_fully_qualified_id(model: Model) -> str:
    return f"{model.metadata.schema}.{model.__tablename__}"


def get_schema(table_id: str) -> Schema:
    """
    Given a `table_id`, returns the `Schema` of the table by querying BigQueries API
    """
    client = get_client()
    table = client.get_table(table_id)
    return table.schema


def column_for_schema_field(schema: SchemaField, **kwargs) -> Column:
    """
    Build a `Column` from a `google.cloud.bigquery.schema.SchemaField`

    Read more: https://cloud.google.com/bigquery/docs/reference/rest/v2/tables#tablefieldschema
    """
    if schema.mode == "REPEATED":
        if schema.field_type == "RECORD":
            column_type = ARRAY(struct_for_schema_field(schema))
        else:
            column_type = ARRAY(BIGQUERY_TYPES_TO_SQLALCHEMY_TYPES[schema.field_type])
    else:
        if schema.field_type == "RECORD":
            column_type = struct_for_schema_field(schema)
        else:
            column_type = BIGQUERY_TYPES_TO_SQLALCHEMY_TYPES[schema.field_type]

    return Column(
        name=schema.name, type_=column_type, comment=schema.description, **kwargs
    )


def schema_for_struct(struct: STRUCT) -> Schema:
    return [
        SchemaField(
            name=name,
            field_type=SQLALCHEMY_TYPES_TO_BIGQUERY_TYPES[sqla_type.__class__],
        )
        for name, sqla_type in struct._STRUCT_fields
    ]


def schema_field_for_column(column: Column) -> SchemaField:
    fields: Optional[SchemaField] = None

    if isinstance(column.type, ARRAY):
        mode = "REPEATED"
        item_type = column.type.item_type
        field_type = SQLALCHEMY_TYPES_TO_BIGQUERY_TYPES[column.type.item_type.__class__]

        if isinstance(item_type, STRUCT):
            fields = tuple(schema_for_struct(item_type))
    else:
        mode = "NULLABLE"
        column_type = column.type
        field_type = SQLALCHEMY_TYPES_TO_BIGQUERY_TYPES[column_type.__class__]

        if isinstance(column_type, STRUCT):
            fields = tuple(schema_for_struct(column_type))

    return SchemaField(
        name=column.name, field_type=field_type, mode=mode, fields=fields or ()
    )


def schema_for_model(model: Model) -> Schema:
    """
    Given an `AmoraModel`, returns the equivalent bigquery `Schema`
    of the model by parsing the model SQLAlchemy column schema
    """
    columns = model.__table__.columns
    return [schema_field_for_column(col) for col in columns]


def schema_for_model_source(model: Model) -> Optional[Schema]:
    """
    Give an `Amora Model`, returns the bigquery `Schema` of its
    `source` classmethod query result
    """
    source = model.source()
    if source is None:
        return None

    result = dry_run(model)
    if result is None:
        return None

    return result.schema


@logger.log_execution()
def run(statement: Compilable) -> RunResult:
    """
    Executes a given query and returns its results
    and metadata as an `amora.providers.bigquery.RunResult`
    """
    query = compile_statement(statement)
    query_job = get_client().query(query)
    rows = query_job.result()
    execution_time_delta = query_job.ended - query_job.started

    return RunResult(
        execution_time_in_ms=execution_time_delta.microseconds / 1000,
        job_id=query_job.job_id,
        query=query,
        referenced_tables=[
            ".".join(table.to_api_repr().values())
            for table in query_job.referenced_tables
        ],
        rows=rows,
        schema=query_job.schema,
        total_bytes=query_job.total_bytes_billed,
        user_email=query_job.user_email,
        to_dataframe=query_job.to_dataframe,
    )


@logger.log_execution()
def dry_run(model: Model) -> Optional[DryRunResult]:
    """
    You can use the estimate returned by the dry run to calculate query
    costs in the pricing calculator. Also useful to verify user permissions
    and query validity. You are not charged for performing the dry run.

    Read more: https://cloud.google.com/bigquery/docs/dry-run-queries

    E.g:
    ```python
    dry_run(HeartRate)
    ```

    Will result in:

    ```python
    DryRunResult(
        total_bytes_processed=170181834,
        query=\"SELECT\\n  `health`.`creationDate`,\\n  `health`.`device`,\\n  `health`.`endDate`,\\n  `health`.`id`,\\n  `health`.`sourceName`,\\n  `health`.`startDate`,\\n  `health`.`unit`,\\n  `health`.`value`\\nFROM `diogo`.`health`\\nWHERE `health`.`type` = 'HeartRate'\",
        model=HeartRate,
        referenced_tables=["amora-data-build-tool.diogo.health"],
        schema=[
            SchemaField("creationDate", "TIMESTAMP", "NULLABLE", None, (), None),
            SchemaField("device", "STRING", "NULLABLE", None, (), None),
            SchemaField("endDate", "TIMESTAMP", "NULLABLE", None, (), None),
            SchemaField("id", "INTEGER", "NULLABLE", None, (), None),
            SchemaField("sourceName", "STRING", "NULLABLE", None, (), None),
            SchemaField("startDate", "TIMESTAMP", "NULLABLE", None, (), None),
            SchemaField("unit", "STRING", "NULLABLE", None, (), None),
            SchemaField("value", "FLOAT", "NULLABLE", None, (), None),
        ],
    )
    ```
    """
    client = get_client()
    source = model.source()
    if source is None:
        table = client.get_table(get_fully_qualified_id(model))

        if table.table_type == "VIEW":
            query_job = client.query(
                query=table.view_query,
                job_config=QueryJobConfig(dry_run=True, use_query_cache=False),
            )
            return DryRunResult(
                job_id=query_job.job_id,
                model=model,
                query=table.view_query,
                referenced_tables=[
                    ".".join(table.to_api_repr().values())
                    for table in query_job.referenced_tables
                ],
                schema=query_job.schema,
                total_bytes=query_job.total_bytes_processed,
                user_email=query_job.user_email,
            )

        return DryRunResult(
            job_id=None,
            model=model,
            query=None,
            referenced_tables=[str(table.reference)],
            schema=table.schema,
            total_bytes=table.num_bytes,
            user_email=None,
        )

    query = compile_statement(source)

    query_job = client.query(
        query=query,
        job_config=QueryJobConfig(dry_run=True, use_query_cache=False),
    )
    return DryRunResult(
        job_id=query_job.job_id,
        total_bytes=query_job.total_bytes_processed,
        referenced_tables=[
            ".".join(table.to_api_repr().values())
            for table in query_job.referenced_tables
        ],
        query=query,
        model=model,
        schema=query_job.schema,
        user_email=query_job.user_email,
    )


class fixed_unnest(sqlalchemy.sql.roles.InElementRole, unnest):
    _with_offset = None

    def __init__(self, *args, **kwargs):
        self.name = "unnest"
        super().__init__(*args, **kwargs)

    def table_valued(self, *expr, with_offset: str = None, **kwargs):
        new_func = self._generate()

        if with_offset:
            expr += (with_offset,)
            new_func._with_offset = with_offset

        new_func.type = new_func._table_value_type = sqltypes.TableValueType(*expr)

        return new_func


def cte_from_rows(rows: Iterable[Dict[str, Any]]) -> CTE:
    """
    Returns a table like selectable (CTE) for the given hardcoded values.

    E.g:
    ```python
    rows = [{"numeric_column": "123"}, {"numeric_column": "234"}, {"numeric_column": "345"}]
    cte_from_rows(rows)
    ```

    Will result in the following SQL

    ```sql
        WITH `annon_1` AS (
            SELECT "123" AS numeric_column
            UNION ALL SELECT "234 AS numeric_column
            UNION ALL SELECT "345" AS numeric_column
        )
    ```

    Which would render a table like:

    ```md
    | numeric_column |
    |----------------|
    | 123            |
    | 234            |
    | 345            |
    ```

    Useful both for model writing and testing purposes. Think of `cte_from_rows` as way of generating
    a "temporary table like object", with data available at runtime.

    """

    def gen_selects(rows):
        for row in rows:
            cols = []
            for name, value in row.items():
                if isinstance(value, array):
                    cols.append(value.label(name))
                elif isinstance(value, AmoraModel):
                    cols.append(struct(value).label(name))
                else:
                    cols.append(literal(value).label(name))
            yield select(cols)

    selects = list(gen_selects(rows))

    if len(selects) == 1:
        return selects[0].cte()

    return selects[0].union_all(*(selects[1:])).cte()


def estimated_query_cost_in_usd(total_bytes: int) -> float:
    """
    By default, queries are billed using the on-demand pricing model,
    where you pay for the data scanned by your queries.

    - This function doesn't take into consideration that the first 1 TB per month is free.
    - By default, the estimation is based on BigQuery's `On-demand analysis` pricing, which may change over time and
    may vary according to regions and your personal contract with GCP.

    You may set `AMORA_GCP_BIGQUERY_ON_DEMAND_COST_PER_TERABYTE_IN_USD` to the appropriate value for your use case.

    More on: https://cloud.google.com/bigquery/pricing#analysis_pricing_models

    :param total_bytes: Total data processed by the query
    :return: The estimated cost in USD, based on `On-demand` price
    """
    total_terabytes = total_bytes / 1024**4
    return total_terabytes * settings.GCP_BIGQUERY_ON_DEMAND_COST_PER_TERABYTE_IN_USD


def estimated_storage_cost_in_usd(total_bytes: int) -> float:
    """
    Storage pricing is the cost to store data that you load into BigQuery.
    `Active storage` includes any table or table partition that has been modified in the last 90 days.

    - This function doesn't take into consideration that the first 10 GB of storage per month is free.
    - By default, the estimation is based on BigQuery's `Active Storage` cost per GB, which may change over time and
    may vary according to regions and your personal contract with GCP.

    You may set `AMORA_GCP_BIGQUERY_ACTIVE_STORAGE_COST_PER_GIGABYTE_IN_USD` to the appropriate value for your use case.

    More on: https://cloud.google.com/bigquery/pricing#storage

    :param total_bytes: Total bytes stored into the table
    :return: The estimated cost in USD, based on `Active storage` price
    """
    total_gigabytes = total_bytes / 1024**3
    return (
        total_gigabytes * settings.GCP_BIGQUERY_ACTIVE_STORAGE_COST_PER_GIGABYTE_IN_USD
    )


def struct_for_model(model: Union[Model, AmoraModel]) -> STRUCT:
    """
    Build a BigQuery Struct type from an AmoraModel specification

    Read more: https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types#struct_type
    """

    def fields():
        for field in model.__fields__.values():
            if field.shape is SHAPE_LIST:
                if issubclass(field.type_, AmoraModel):
                    yield field.name, ARRAY(struct_for_model(field.type_))
                else:
                    yield field.name, ARRAY(get_sqlachemy_type(field))
            elif field.sub_fields is None:
                if issubclass(field.type_, AmoraModel):
                    yield field.name, struct_for_model(field.type_)
                else:
                    yield field.name, get_sqlachemy_type(field)
            else:
                raise NotImplementedError

    return STRUCT(*fields())


def struct_for_schema_field(schema_field: SchemaField) -> STRUCT:
    """
    Build a BigQuery Struct type from a `google.cloud.bigquery.schema.SchemaField`

    Read more: https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types#struct_type
    """

    def fields():
        for field in schema_field.fields:
            if field.mode == "REPEATED":
                if field.field_type == "RECORD":
                    yield field.name, ARRAY(struct_for_schema_field(field))
                else:
                    sqla_type = BIGQUERY_TYPES_TO_SQLALCHEMY_TYPES[field.field_type]
                    yield field.name, ARRAY(sqla_type)
            else:
                if field.field_type == "RECORD":
                    yield field.name, struct_for_schema_field(field)
                else:
                    sqla_type = BIGQUERY_TYPES_TO_SQLALCHEMY_TYPES[field.field_type]
                    yield field.name, sqla_type

    return STRUCT(*fields())


class struct(expression.ClauseList, expression.ColumnElement):  # type: ignore
    """
    A BigQuery STRUCT/RECORD literal.

    !!! warning "Experimental feature. You should probably use struct_for_model"
    """

    __visit_name__ = "struct"

    def __init__(self, model: AmoraModel, **kw):
        self._model = model
        self.type = struct_for_model(model)
        clauses = [
            coercions.expect(
                roles.ExpressionElementRole,
                model.dict(),
                type_=self.type,
            )
        ]
        super().__init__(*clauses, **kw)

    def bind_expression(self, bindvalue):
        return bindvalue

    def self_group(self, against=None):
        if against in (operators.any_op, operators.all_op, operators.getitem):
            return expression.Grouping(self)

        return self


class array(expression.ClauseList, expression.ColumnElement):  # type: ignore
    """
    A BigQuery ARRAY literal.

    This is used to produce `ARRAY` literals in SQL expressions, e.g.:

    ```python
    from sqlalchemy import select

    from amora.compilation import compile_statement
    from amora.providers.bigquery import array

    stmt = select([array([1, 2]).label("a"), array([3, 4, 5]).label("b")])

    compile_statement(stmt)
    ```
    Produces the SQL:

    ```sql
    SELECT
        ARRAY[1, 2] AS a,
        ARRAY[3, 4, 5]) AS b
    ```

    An instance of `array` will always have the datatype `sqlalchemy_bigquery.base.BQArray`.
    The "inner" type of the array is inferred from the values present, unless the
    ``type_`` keyword argument is passed, e.g.:

    ```python
    array(["foo", "bar"], type_=String)
    ```

    Arrays can also be constructed using `AmoraModel` instances,
    which would compile into a array of structs. E.g:

    ```python
    class Point(AmoraModel):
        x: int
        y: int


    array([Point(x=4, y=4), Point(x=2, y=2)])
    ```
    """

    __visit_name__ = "array"

    def __init__(self, clauses, **kw):
        if clauses and isinstance(clauses[0], AmoraModel):
            clauses = [struct(c) for c in clauses]

        clauses = [coercions.expect(roles.ExpressionElementRole, c) for c in clauses]
        super().__init__(*clauses, **kw)
        self._type_tuple = [arg.type for arg in clauses]
        main_type = kw.pop(
            "type_",
            self._type_tuple[0] if self._type_tuple else sqltypes.NULLTYPE,
        )
        self.type = BQArray(main_type, dimensions=1)

        for type_ in self._type_tuple:
            if type(type_) is sqltypes.NullType:
                raise ValueError("Array cannot have a null element")

    def _bind_param(self, operator, obj, _assume_scalar=False, type_=None):
        if _assume_scalar or operator is operators.getitem:
            # if getitem->slice were called, Indexable produces
            # a Slice object from that
            assert isinstance(obj, int)
            return expression.BindParameter(
                None,
                obj,
                _compared_to_operator=operator,
                type_=type_,
                _compared_to_type=self.type,
                unique=True,
            )

        return array(
            [
                self._bind_param(operator, o, _assume_scalar=True, type_=type_)
                for o in obj
            ]
        )

    def self_group(self, against=None):
        if against in (operators.any_op, operators.all_op, operators.getitem):
            return expression.Grouping(self)

        return self


def zip_arrays(
    *arr_columns: Column, additional_columns: Optional[List[Column]] = None
) -> Compilable:
    """
    Given at least two array columns of equal length, returns a table of the unnest values,
    converting array items into rows. E.g:

    A CTE with 3 array columns: `entity`, `f1`, `f2`

    ```python
    from amora.providers.bigquery import array, cte_from_rows, zip_arrays

    cte = cte_from_rows(
        [
            {
                "entity": array([1, 2]),
                "f1": array(["f1v1", "f1v2"]),
                "f2": array(["f2v1", "f2v2"]),
            }
        ]
    )

    zip_arrays(cte.c.entity, cte.c.f1, cte.c.f2)
    ```

    Will result in the following table:

    | entity | f1   | f2   |
    |--------|------|------|
    | 1      | f1v1 | f2v1 |
    | 2      | f1v2 | f2v2 |

    If additional columns are needed from the original data, those can be selected
    using the optional `additional_columns`:

    ```python
    from amora.providers.bigquery import array, cte_from_rows, zip_arrays

    cte = cte_from_rows(
        [
            {
                "entity": array([1, 2]),
                "f1": array(["f1v1", "f1v2"]),
                "f2": array(["f2v1", "f2v2"]),
                "id": 1,
            },
            {
                "entity": array([3, 4]),
                "f1": array(["f1v3", "f1v4"]),
                "f2": array(["f2v3", "f2v4"]),
                "id": 2,
            },
        ]
    )

    zip_arrays(cte.c.entity, cte.c.f1, cte.c.f2, additional_columns=[cte.c.id])
    ```

    | entity | f1   | f2   | id   |
    |--------|------|------|------|
    | 1      | f1v1 | f2v1 | 1    |
    | 2      | f1v2 | f2v2 | 1    |
    | 3      | f1v3 | f2v3 | 2    |
    | 4      | f1v4 | f2v4 | 2    |

    Read more: [https://cloud.google.com/bigquery/docs/reference/standard-sql/arrays#zipping_arrays](https://cloud.google.com/bigquery/docs/reference/standard-sql/arrays#zipping_arrays)

    Args:
        *arr_columns: Array columns of equal length
        additional_columns: Additional columns needed from the original data
    """
    offset_alias = "off"
    offset = func.offset(literal_column(offset_alias))

    columns: List[ColumnElement] = [col[offset].label(col.key) for col in arr_columns]
    if additional_columns:
        columns += additional_columns

    return select(columns).join(
        fixed_unnest(arr_columns[0]).table_valued(with_offset=offset_alias),
        onclause=literal(1) == literal(1),
        isouter=True,
    )


def _sample_cache_key(
    model, percentage=1, limit=settings.GCP_BIGQUERY_DEFAULT_LIMIT_SIZE
) -> str:
    return f"{model.unique_name()}.{percentage}.{limit}.{date.today()}"


@cache(_sample_cache_key)
def sample(
    model: Model,
    percentage: int = 1,
    limit: int = settings.GCP_BIGQUERY_DEFAULT_LIMIT_SIZE,
) -> pd.DataFrame:
    """
    Given a model, returns a random sample of the data.

    Read more: [https://cloud.google.com/bigquery/docs/table-sampling](https://cloud.google.com/bigquery/docs/table-sampling)



    Args:
        model: AmoraModel to extract the extract the sample
        percentage: The percentage of the sample. E.g: percentage=10 is 10% of the data.
        limit: The maximum number of rows to be returned
    Returns:
        The sample data as a `pandas.DataFrame`
    Raises:
        ValueError: TABLESAMPLE SYSTEM can only be applied directly to tables.
    """
    if model.__model_config__.materialized is not MaterializationTypes.table:
        raise ValueError(
            "TABLESAMPLE SYSTEM can only be applied directly to tables. "
            "More on: https://cloud.google.com/bigquery/docs/table-sampling#limitations"
        )

    sampling = literal_column(f"{percentage} PERCENT")
    model_sample = tablesample(model, sampling)  # type: ignore
    stmt = select(model_sample).limit(limit)

    return run(stmt).to_dataframe()



import logging
import sys
import time
from datetime import timedelta
from functools import wraps
from typing import Callable

import humanize

from amora.config import settings

logger = logging.getLogger("amora")
logger.setLevel(settings.LOGGER_LOG_LEVEL)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))


def log_execution():
    """
    Wraps the execution of a function, calculates the execution time and logs the
    result in a human readable way. E.g:

    ```python
    from time import sleep

    from amora import logger


    @logger.log_execution()
    def do_something():
        sleep(3)
        return 42


    do_something()
    ```

    Would result in the log
    ```
    Function call to `a_module.do_something` took 3 seconds
    ```

    """

    def wrapper(fn: Callable):
        @wraps(fn)
        def decorator(*args, **kwargs):
            t0 = time.perf_counter()

            result = fn(*args, **kwargs)

            execution_delta = timedelta(seconds=time.perf_counter() - t0)
            f_name = f"{fn.__module__}.{fn.__qualname__}"
            delta = humanize.naturaldelta(execution_delta, minimum_unit="milliseconds")

            logger.debug("Function call to `%s` took %s", f_name, delta)

            return result

        return decorator

    return wrapper



import logging
import os
from enum import Enum
from pathlib import Path
from tempfile import NamedTemporaryFile, mkdtemp
from typing import Optional, Tuple
from uuid import uuid4

from pydantic import BaseSettings, root_validator, validator

ROOT_PATH = Path(__file__).parent.parent
AMORA_MODULE_PATH = ROOT_PATH.joinpath("amora")

_Width = float
_Height = float


class StorageCacheProviders(str, Enum):
    local = "local"
    gcs = "gcs"


class Settings(BaseSettings):
    TARGET_PROJECT: str
    TARGET_SCHEMA: str

    PROJECT_PATH: Path
    DASHBOARDS_PATH: Optional[Path]
    MODELS_PATH: Optional[Path]
    TARGET_PATH: Optional[Path]

    CLI_CONSOLE_MAX_WIDTH: int = 160
    CLI_MATERIALIZATION_DAG_FIGURE_SIZE: Tuple[_Width, _Height] = (32, 32)

    # https://cloud.google.com/bigquery/pricing#analysis_pricing_models
    GCP_BIGQUERY_ON_DEMAND_COST_PER_TERABYTE_IN_USD: float = 5.0
    # https://cloud.google.com/bigquery/pricing#storage
    GCP_BIGQUERY_ACTIVE_STORAGE_COST_PER_GIGABYTE_IN_USD: float = 0.020

    GCP_BIGQUERY_DEFAULT_LIMIT_SIZE: int = 1000

    LOCAL_ENGINE_ECHO: bool = False
    LOCAL_ENGINE_SQLITE_FILE_PATH: Path = Path(
        NamedTemporaryFile(suffix="amora-sqlite.db", delete=False).name
    )
    STORAGE_CACHE_ENABLED: bool = False
    STORAGE_CACHE_PROVIDER: StorageCacheProviders = StorageCacheProviders.local
    STORAGE_GCS_BUCKET_NAME: str = "amora-storage"
    STORAGE_LOCAL_CACHE_PATH: Path = Path(mkdtemp())
    STORAGE_PARQUET_ENGINE: str = "pyarrow"

    LOGGER_LOG_LEVEL: int = logging.DEBUG

    MONEY_DECIMAL_PLACES: int = 4

    TEST_RUN_ID: str = os.getenv("PYTEST_XDIST_TESTRUNUID") or f"amora-{uuid4().hex}"

    class Config:
        env_prefix = "AMORA_"

    @root_validator
    def compute_MODELS_PATH(cls, values: dict) -> dict:
        if values["MODELS_PATH"] is not None:
            return values

        values["MODELS_PATH"] = values["PROJECT_PATH"].joinpath("models")
        return values

    @root_validator
    def compute_DASHBOARDS_PATH(cls, values: dict) -> dict:
        if values["DASHBOARDS_PATH"] is not None:
            return values

        values["DASHBOARDS_PATH"] = values["PROJECT_PATH"].joinpath("dashboards")
        return values

    @root_validator
    def compute_TARGET_PATH(cls, values: dict) -> dict:
        if values["TARGET_PATH"] is not None:
            return values

        target_path: Path = values["PROJECT_PATH"].joinpath(".target")
        if not target_path.exists():
            target_path.mkdir(exist_ok=False)

        values["TARGET_PATH"] = target_path

        return values

    @validator("PROJECT_PATH")
    def project_path_is_a_valid_path(cls, v: Path) -> Path:
        if not v.is_dir():
            raise ValueError(f"{v.as_posix()} must be a valid directory")
        return v

    @property
    def models_path(self) -> Path:
        assert isinstance(self.MODELS_PATH, Path)
        return self.MODELS_PATH

    @property
    def target_path(self) -> Path:
        assert isinstance(self.TARGET_PATH, Path)
        return self.TARGET_PATH

    @property
    def dashboards_path(self) -> Path:
        assert isinstance(self.DASHBOARDS_PATH, Path)
        return self.DASHBOARDS_PATH


settings = Settings()



from dash import dash_table, html

from amora.models import Model
from amora.providers.bigquery import sample


def component(model: Model) -> dash_table.DataTable:
    try:
        df = sample(model, percentage=1)
    except ValueError:
        return html.Div(f"Sample not implemented for model {model.unique_name()}")
    else:
        return dash_table.DataTable(
            data=df.to_dict("records"),
            columns=[{"name": col, "id": col} for col in sorted(df.columns.values)],
            export_format="csv",
            sort_action="native",
        )



import sqlmodel
from feast import ValueType
from sqlalchemy.sql import sqltypes

from amora.models import Column

_SQLALCHEMY_TYPES_TO_FS_TYPES = {
    sqltypes.Float: ValueType.FLOAT,
    sqltypes.String: ValueType.STRING,
    sqlmodel.AutoString: ValueType.STRING,
    sqltypes.Integer: ValueType.INT64,
    sqltypes.Boolean: ValueType.BOOL,
    sqltypes.TIMESTAMP: ValueType.UNIX_TIMESTAMP,
    (sqltypes.ARRAY, sqltypes.String): ValueType.STRING_LIST,
    (sqltypes.ARRAY, sqltypes.Integer): ValueType.INT64_LIST,
    (sqltypes.ARRAY, sqltypes.Boolean): ValueType.BOOL_LIST,
    (sqltypes.ARRAY, sqltypes.Float): ValueType.FLOAT_LIST,
}


def type_for_column(col: Column) -> ValueType:
    if isinstance(col.type, sqltypes.ARRAY):
        key = (col.type.__class__, col.type.item_type.__class__)
    else:
        key = col.type.__class__

    return _SQLALCHEMY_TYPES_TO_FS_TYPES[key]



from urllib.parse import quote_plus, urlencode

from authlib.integrations.flask_client import OAuth
from dash import Dash
from flask import redirect, session, url_for

from amora.dash.config import settings


def add_auth0_login(dash_app: Dash) -> None:
    flask_app = dash_app.server
    flask_app.secret_key = settings.APP_SECRET_KEY.get_secret_value()
    oauth = OAuth(flask_app)

    oauth.register(
        "auth0",
        client_id=settings.AUTH0_CLIENT_ID,
        client_secret=settings.AUTH0_CLIENT_SECRET.get_secret_value(),  # type: ignore
        client_kwargs={
            "scope": settings.AUTH0_SCOPE,
        },
        server_metadata_url=f"https://{settings.AUTH0_DOMAIN}/.well-known/openid-configuration",
    )

    @flask_app.route(settings.AUTH0_LOGIN_PATH)
    def login():
        return oauth.auth0.authorize_redirect(
            redirect_uri=url_for("callback", _external=True)
        )

    @flask_app.route(settings.AUTH0_CALLBACK_PATH, methods=["GET", "POST"])
    def callback():
        token = oauth.auth0.authorize_access_token()
        session["user"] = token
        return redirect(settings.AUTH0_CALLBACK_REDIRECT_PATH)

    @flask_app.route(settings.AUTH0_LOGOUT_PATH)
    def logout():
        session.clear()
        return redirect(
            "https://"
            + settings.AUTH0_DOMAIN
            + "/v2/logout?"
            + urlencode(
                {
                    "returnTo": url_for(
                        settings.AUTH0_CALLBACK_REDIRECT_PATH, _external=True
                    ),
                    "client_id": settings.AUTH0_CLIENT_ID,
                },
                quote_via=quote_plus,
            )
        )
